"""
This module contains the Experiment class and any future add ons.

Created on Tue Dec 21 08:30:33 2021.
@author: Briley Bourgeois
"""
import re
import os
import time
from ast import literal_eval
from datetime import date

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


class Experiment:
    """
    Object containing all information necessary to run a particular experiment.

    The experiment class is the center piece of catalight. This class
    contains several type checked properties that should be updated with
    relevant experimental parameters by the user. The user sets a desired
    experimental procedure from the list provided in :py:attr:`expt_list`

    Most importantly, equipment object can be passed to each instance of
    Experiment and used to run the :py:meth:`run_experiment` method to control
    the hardware components.

    Parameters
    ----------
    eqpt_list: list of objects, optional
        List of equipment objects. Calls :py:meth:`update_eqpt_list`,
        if provided.
        Order of list should be:
        (:py:class:`~catalight.equipment.sri_gc.gc_control.GC_Connector`,
        :class:`~catalight.equipment.diode_laser.diode_control.Diode_Laser`,
        :class:`~catalight.equipment.alicat_MFC.gas_control.Gas_System`,
        :class:`~catalight.equipment.harrick_watlow.heater_control.Heater`)
    """

    def __init__(self, eqpt_list=False):
        """Init experiment object."""
        # This class attribute defines the possible experiments. This is an
        # important part of the class and should be altered with caution
        # noqa to suppress extra spaces PEP violation. Done for clarity.

        # TODO: ind_var of stability test should be changed to time, but
        # careful bug checking needs to be performed everywhere expt.ind_var
        # is used.
        self._expt_list = pd.DataFrame(
            [['temp_sweep',      'temp', False,    'K'],  # noqa
             ['power_sweep',    'power', False,   'mW'],  # noqa
             ['comp_sweep',  'gas_comp', False, 'frac'],  # noqa
             ['flow_sweep',  'tot_flow', False, 'sccm'],  # noqa
             ['calibration', 'gas_comp', False,  'ppm'],  # noqa
             ['stability_test',  'temp', False,  'min']], # noqa
            columns=['Expt Name',
                     'Independent Variable',
                     'Active Status',
                     'Units'])

        # Set default values of properties
        # --------------------------------
        # These are just random default values.
        # Descriptions below in property definitions
        self._temp = [273.0]
        self._power = [0.0]
        self._tot_flow = [0.0]
        self._gas_comp = [[0.0, 50.0, 0.0, 0.0]]
        self._gas_type = ['C2H2', 'Ar', 'H2', 'Ar']
        self._sample_rate = 10

        # Returns todays date as YYYYMMDD by default
        self._date = date.today().strftime('%Y%m%d')
        self._start_time = 'Undefined'
        self._expt_type = 'Undefined'
        self._sample_name = 'Undefined'
        self._ind_var = 'Undefined'
        self._expt_name = 'Undefined'
        self._results_path = 'Undefined'
        self._data_path = 'Undefined'

        # Define attributes and set default
        # ---------------------------------
        self.sample_set_size = 4
        """int: Number of GC samples to collect"""

        self.t_buffer = 5
        """
        int or float: (min) Time between last GC sample of a step and going to
        next condition
        """

        self.t_steady_state = 15
        """
        int or float: (min) Time between going to next condition and collecting
        first sample of a step
        """

        self.heat_rate = 15
        """int or float: (deg C/min) Ramp rate to use when heating reactor."""

        if eqpt_list is not False:
            self.update_eqpt_list(eqpt_list)

    # These setter functions apply rules for how certain properties can be set
    def _str_setter(attr):
        """
        Setter function for properties with str values.

        If type checking becomes necessary for string inputs,
        it can be done here

        Parameters
        ----------
        attr : str
            str for non-public attribute to change value of

        Returns
        -------
        set_any : function
            update function for setting string properties
        """

        def set_any(self, value):
            """
            Func to set value.

            Parameters
            ----------
            value : str
                value to attempt update

            Returns
            -------
            None.

            """
            setattr(self, attr, value)

        return set_any

    def _num_setter(attr):
        """
        Setter function for numbers used to set checks on value changes.

        Parameters
        ----------
        attr : `str`
            str for non-public attribute to change value of

        Raises
        ------
        AttributeError
            Attribute error is raised when value entered doesn't match with
            corresponding limits. These vary depending on attribute given.

        Returns
        -------
        set_any : function
            update function for setting string properties

        """

        def set_any(self, value):
            """
            Setter function for numbers.

            Converts np.ndarrays to list on call. Runs through value testing if
            statements based on attr supplied. All values provided must be a
            form of a list. Gas comp values must individually round to 1.
            Total flow is currently hard coded to max at 350.

            Parameters
            ----------
            value : list of float or numpy.ndarray
                if np array is given, converts to list in beginning of func.

            Raises
            ------
            AttributeError
                Attribute error is raised when value entered doesn't match with
                corresponding limits. These vary depending on attribute given.
            """
            if isinstance(value, np.ndarray):
                value = list(value)

            if not isinstance(value, list):
                raise AttributeError(attr + ' must be list')
            elif (attr == '_tot_flow') & (np.max(value) > 350):
                raise AttributeError('Total flow must be <= 350')
                # TODO, update this value if gas_system is connected or turn
                # into a class attribute.
            elif (attr == '_gas_comp'):
                for composition in value:
                    # convert % to fraction
                    if round(sum(composition), 6) == 100:
                        composition[:] = [x / 100 for x in composition]

                    if round(sum(composition), 6) != 1:
                        raise AttributeError(
                            'Gas comp. must be list of list == 1')

            setattr(self, attr, value)
        return set_any

    def _attr_getter(attr):
        """Getter function for all properties."""

        def get_any(self):
            return getattr(self, attr)
        return get_any

    # Property definitions
    # ====================
    # Number Properties
    # -----------------
    temp = property(fget=_attr_getter('_temp'),
                    fset=_num_setter('_temp'))
    """
    list[float]: List of temperature values to step through during experiment.
    One element if constant or multiple for sweep. Initial value of [273.0].
    """
    power = property(fget=_attr_getter('_power'),
                     fset=_num_setter('_power'))
    """
    list[float]: List of power (mW) values to step through during experiment.
    One element if constant or multiple for sweep. Initial value of [0.0].
    """

    tot_flow = property(fget=_attr_getter('_tot_flow'),
                        fset=_num_setter('_tot_flow'))
    """
    list[float]: List of flow rates (sccm) to step through during experiment.
    One element if constant or multiple for sweep. Initial value of [0.0].
    """

    gas_comp = property(fget=_attr_getter('_gas_comp'),
                        fset=_num_setter('_gas_comp'))
    """
    list[list[float]]: List of gas composition lists to sweep through.
    (ex. [[gas1, gas2, gas3],[...]]) Can be len=1 if not comp sweep.
    Each inner list must sum to 1 or 100.
    Initial value of  [[0.0, 50.0, 0.0, 0.0]].
    """

    # String Properties
    # -----------------
    gas_type = property(fget=_attr_getter('_gas_type'),
                        fset=_str_setter('_gas_type'))
    """
    list[str]: List of gas types for the MFCs to use to measure flow rates.
    (ex. [gasA, gasB, gasC, ...]) All values must be found in
    the gas library of the MFC system used.
    Initial value of ['C2H2', 'Ar', 'H2', 'Ar'].
    """

    sample_name = property(fget=_attr_getter('_sample_name'),
                           fset=_str_setter('_sample_name'))
    """str: Name of sample for building save paths"""

    # Read only properties
    # --------------------
    date = property(lambda self: self._date)
    """str, read-only: Update w/ :meth:`update_date` method. For logging"""

    start_time = property(lambda self: self._start_time)
    """
    str, read-only: Updates when :meth:`set_initial_conditions` method is
    called. For logging. Used to calculate time_passed during analysis.
    Given by time.time() at the **end** of initial conditions steps.
    """

    ind_var = property(lambda self: self._ind_var)
    """
    str, read-only: Describes the variable being modified. This gets updated
    when expt_type is updated based on what defined by :attr:`expt_list`.
    """

    expt_name = property(lambda self: self._expt_name)
    """str, read-only: Creates name using the fixed variables for expt"""

    results_path = property(lambda self: self._results_path)
    """
    str, read-only: Save location for analysis, defined relative to expt_log.
    Update using :meth:`update_save_paths`
    """

    data_path = property(lambda self: self._data_path)
    """
    str, read-only: Save location for raw data, defined relative to expt_log.
    Update using :meth:`update_save_paths`
    """

    expt_list = property(lambda self: self._expt_list)
    """
    pandas.DataFrame, read-only: This class attr defines the possible
    experiments. This is an important part of the class and should be altered
    with caution. Changing the units within this DF should allow different unit
    inputs for the rest of the codebase, but this feature is untested!!!
    """

    # Setting sample rate changes when connected to GC
    @property
    def sample_rate(self):
        """
        Sample rate in minutes (time it takes to collect a sample).

        Set sample rate based on whether GC is connected.
        If GC is connected, try to set sample rate to that defined by expt.
        If that failed, _gc_control.sample_rate will be set to min defined by
        ctrl file.
        """
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        """Returns sample rate value."""
        if hasattr(self, '_gc_control'):
            if value >= self._gc_control.min_sample_rate:
                # try to set sample rate to that defined by expt
                self._sample_rate = value
                self._gc_control.sample_rate = self.sample_rate
            else:
                # If that failed, _gc_control.sample_rate will be set to min
                # defined by ctrl file. This line makes sure gc and expt match.
                self._sample_rate = self._gc_control.sample_rate
                print('Error: Input Sample Rate is Lower Than GC Minimum')
                print('Sample rate set to: %.2f'
                      % self._gc_control.sample_rate)

        else:
            self._sample_rate = value

    @property
    def expt_type(self):
        """
        Define the desired experimental procedure selected from _expt_list.

        A key property directing much of the behavior of the class.For example,
        the run_experiment() method is guided by the expt_type provided. New
        values must be entered into the _expt_list attribute and the
        appropriate class methods must be edited to account for new
        experimental capabilities.
        Updates :attr:`ind_var` when called.
        Converts 'Active Status' in :attr:`expt_list` to True.
        """
        return self._expt_type

    @expt_type.setter
    def expt_type(self, expt_type):
        # Evaluate whether entered experiment type is in default list
        if expt_type not in self.expt_list['Expt Name'].tolist():
            raise AttributeError('Invalid Experiment Type Provided')

        self._expt_type = expt_type
        self._expt_list['Active Status'] = (
            self.expt_list['Expt Name'] == self._expt_type)

        # Defines Independent Variable as string provided by expt list
        var_list = self.expt_list['Independent Variable']  # Pull List
        # Compare Boolean
        ind_var_series = var_list[self.expt_list['Active Status']]
        # Convert to str
        self._ind_var = ind_var_series.to_string(index=False)

    def update_date(self):
        """Set :attr:`date` based on current date."""
        self._date = date.today().strftime('%Y%m%d')

    def update_expt_log(self, expt_path):
        """
        Update the experiment log based on current object parameters.

        Parameters
        ----------
        expt_path: str
            path to the experiment folder
        """
        with open(os.path.join(expt_path, 'expt_log.txt'), 'w+') as log:
            log_entry = [
                'Experiment Date = ' + self.date,
                'Experiment Start = ' + str(self.start_time),
                'Experiment Type = ' + self.expt_type,
                'Experiment Name = ' + self.expt_name,
                'Sample Name = ' + self.sample_name,
                'Temperature [' + self.expt_list['Units'][0] + '] = '
                + str(self.temp),
                'Power [' + self.expt_list['Units'][1] + '] = '
                + str(self.power),
                'Gas 1 type = ' + self.gas_type[0],
                'Gas 2 type = ' + self.gas_type[1],
                'Gas 3 type = ' + self.gas_type[2],
                'Gas 4 type = ' + self.gas_type[3],
                'Gas Composition [' + self.expt_list['Units'][2]
                + '] = ' + str(self.gas_comp),
                'Total Flow [' + self.expt_list['Units'][3]
                + '] = ' + str(self.tot_flow)
            ]
            log.write('\n'.join(log_entry))

    def read_expt_log(self, log_path):
        """
        Read data from an existing log file and update object parameters.

        Grows gas_type list if more gasses are found in the expt_log than
        contained in current gas_type list.

        Parameters
        ----------
        log_path : str
            string to the full file path of the log file ('./expt_log.txt').

        Returns
        -------
        None.

        """
        with open(log_path, 'r') as log:
            for line in log:  # Read file line by line
                # Read the values after '=' sign
                data = line.split('=')[-1].strip(' \n')

                # Check the text before '=' to pick which value to assign
                if re.search('Experiment Date =', line):
                    self._date = data
                elif re.search('Experiment Start =', line):
                    self._start_time = literal_eval(data)
                elif re.search('Experiment Type =', line):
                    self.expt_type = data
                elif re.search('Experiment Name =', line):
                    self._expt_name = data
                elif re.search('Sample Name =', line):
                    self.sample_name = data
                elif re.search('Temperature', line):
                    self.temp = literal_eval(data)
                elif re.search('Power', line):
                    self.power = literal_eval(data)
                elif re.search('Gas \d+ type', line):  # \d+ is 1 or more digit
                    # Get left side of '=' sign
                    gas_str = line.split('=')[0].strip(' \n')
                    # Pull number from left side
                    for string in gas_str.split():
                        if string.isdigit():
                            num = int(string)
                    # If the gas_type list is shorter than this entry number,
                    # append extra blank values. Appending blanks vs just
                    # appending the value accounts for if gas # is out of order
                    if len(self.gas_type) < num:
                        for n in range(num - len(self.gas_type)):
                            self.gas_type.append([])
                    self.gas_type[num-1] = data

                elif re.search('Gas Composition', line):
                    self.gas_comp = literal_eval(data)
                elif re.search('Total Flow', line):
                    self.tot_flow = literal_eval(data)

            expt_path = os.path.dirname(log_path)
            # Will throw error if no data folders
            self.update_save_paths(expt_path)

    def _update_expt_name(self):
        """
        Non-public function that updates expt_name based on current settings.

        Returns
        -------
        None.
        """
        # Defines all settings to be included in path name and adds units
        expt_settings = pd.Series(
            [str(self.temp[0]), str(self.power[0]),
             '_'.join([str(m) + n for m, n in zip(self.gas_comp[0],
                                                  self.gas_type)]),
             str(self.tot_flow[0])]) + self.expt_list['Units'][0:4]

        # Only select fixed variable for path name
        fixed_vars = expt_settings[self.expt_list['Independent Variable']
                                   != self.ind_var]
        self._expt_name = '_'.join(fixed_vars.to_list())  # Define expt_name

    def update_save_paths(self, expt_path, should_exist=True):
        """
        Update data/results paths for the object.

        Used when reading expt from log file or creating new expt data set

        Parameters
        ----------
        expt_path : str
            string to the full file path of the experiments dir.
        should_exist : bool, optional
            If updating based on existing log file, set to true.
            The default is True.

        Raises
        ------
        ValueError
            Gives error if should_exist=True and
            no results/data directories exists,

        Returns
        -------
        None.

        """
        # Defines path for saving results
        self._results_path = os.path.join(expt_path, 'Results')

        # Defines path for saving raw data
        self._data_path = os.path.join(expt_path, 'Data')

        # If updating paths from file, paths should exist in log directory
        exists = (os.path.isdir(self.results_path)
                  | os.path.isdir(self.data_path))
        if (should_exist) & (not exists):
            raise ValueError('results/data path doesn''t exist where expected')

    def create_dirs(self, sample_path):
        """
        Create directory for storing experiment data and analysis.

        This function creates a set of directories to store experimental data
        for a given experiment and the results of the analysis. Updates the
        experiment name based on current expt settings and creates/updates the
        experiment log file in the sample path

        Parameters
        ----------
        sample_path: str
            Input the path where the data should be stored (sample folder)

        Raises
        ------
        AttributeError: Throws error if :attr:`expt_type` is undefined

        Returns
        -------
        None.

        """
        if self.expt_type == 'Undefined':
            raise AttributeError(
                'Experiment Type Must Be Defined Before Creating Directories')

        self._update_expt_name()
        expt_path = os.path.join(sample_path, self.date + self.expt_type
                                 + '_' + self.expt_name)

        self.update_save_paths(expt_path, should_exist=False)

        os.makedirs(self.results_path, exist_ok=True)
        step_num = 1
        if self.expt_type == 'stability_test':
            path = os.path.join(self.data_path, '1 Stability_Test')
            os.makedirs(path, exist_ok=True)
            self.update_expt_log(expt_path)
            return

        # Creates subfolders for each step of experiment
        for step in getattr(self, self._ind_var):
            # Compare Boolean
            units = (self.expt_list['Units']
                     [self.expt_list['Active Status']].to_string(index=False))

            if self._ind_var == 'gas_comp':
                step_str = '_'.join(
                    [str(m) + n for m, n in zip(step, self.gas_type)])
                path = os.path.join(self.data_path,
                                    ('%i %s%s' % (step_num, step_str, units)))
            else:
                path = os.path.join(self.data_path,
                                    ('%i %d%s' % (step_num, step, units)))

            os.makedirs(path, exist_ok=True)
            step_num += 1

        self.update_expt_log(expt_path)

    def plot_sweep(self, fig=None):
        """
        Plot the sweep parameter vs time.

        Parameters
        ----------
        fig : matplotlib.pyplot.figure, optional
            Can supply figure object to write plot to it. The default is None.

        Returns
        -------
        matplotlib.pyplot.figure
            figure object for experimental sweep. Two subplots for full and
            zoomed in versions. There is a hidden single subplot in background
            used for making shared axis titles
        matplotlib.pyplot.axis
            ax1 - top plot showing full experimental sweep
        matplotlib.pyplot.axis
            ax2 - bottom plot showing zoomed in version of experimental sweep
        float
            run_time - Calculate total time to run experiment

        """
        # have to get the sample run time from GC
        # Some Plot Defaults
        plt.rcParams['axes.linewidth'] = 2
        plt.rcParams['lines.linewidth'] = 1.5
        plt.rcParams['xtick.major.size'] = 6
        plt.rcParams['xtick.major.width'] = 1.5
        plt.rcParams['ytick.major.size'] = 6
        plt.rcParams['ytick.major.width'] = 1.5
        # plt.rcParams['figure.figsize'] = (6.5, 8)
        plt.rcParams['font.size'] = 14
        plt.rcParams['axes.labelsize'] = 18
        plt.style.use('seaborn-dark')

        sweep_val = getattr(self, self.ind_var)
        selector = self.expt_list['Active Status']
        sweep_title = (self.expt_list['Independent Variable']
                       [selector].to_string(index=False))
        units = self.expt_list['Units'][selector].to_string(index=False)

        # t_batch is the times for gc collection within one step
        t_batch = list(self.t_steady_state
                       + self.sample_rate * np.arange(0, self.sample_set_size))
        # setpoint0 is 0 or list of zeros for gasses
        setpoint0 = [0] * np.size(sweep_val[0])
        if self.expt_type == 'temp_sweep':
            if units == 'K':
                setpoint0 = 293
            elif units == 'C':
                setpoint0 = 20

            delta_T = np.diff(sweep_val)
            # sets time to change temps based on heat_rate
            t_trans = np.append((sweep_val[0] - setpoint0),
                                delta_T) / self.heat_rate
        else:
            t_trans = np.array([0] * len(sweep_val))  # 0 min transition time

        # Define the values for the first step condition
        # These two define time and value setting of reactor
        t_set = np.array([0, t_trans[0], t_batch[-1] + self.t_buffer])
        setpoint = np.vstack((setpoint0, sweep_val[0], sweep_val[0]))
        # These two define time and value GC collects data
        t_sample = np.array(t_batch)
        # sample_val = np.array([sweep_val[0]]*self.sample_set_size)
        sample_val = np.tile(sweep_val[0], (self.sample_set_size, 1))

        # Loop through remaining setpoints
        for step_num in range(1, len(sweep_val)):
            # new sample times: start at last sample,
            # add t_buffer, add new batch
            t_sample = np.append(t_sample,
                                 t_sample[-1] + self.t_buffer + t_batch)
            sample_val = np.concatenate((sample_val,
                                         np.tile(sweep_val[step_num],
                                                 (self.sample_set_size, 1))))
            # add two points to t_set:
            # 1) last set point time + transition time
            # 2) last set point + time of last gc run + buffer
            t_set = np.append(t_set, [t_set[-1] + t_trans[step_num],
                                      t_set[-1] + t_batch[-1] + self.t_buffer])
            setpoint = np.concatenate(
                (setpoint, np.tile(sweep_val[step_num], (2, 1))))

        if fig is None:
            fig = plt.figure()

        fig.clear()
        ax1 = fig.add_subplot(2, 1, 1)
        ax2 = fig.add_subplot(2, 1, 2)

        ax1.plot(t_set, setpoint, '-o')
        ax1.plot(t_sample, sample_val, 'x')
        ylim_max = 1.1 * np.max(sweep_val)
        ylim_min = 0.9 * np.min(sweep_val) - 0.05
        ax1.set_ylim([ylim_min, ylim_max])

        ax2.plot(t_set[0:3], setpoint[0:3], '-o')
        ax2.plot(t_sample[0:self.sample_set_size],
                 sample_val[0:self.sample_set_size], 'x')
        ylim_max = 1.1 * np.max(sample_val[self.sample_set_size - 1])
        ylim_min = 0.9 * np.max(sample_val[0]) - 0.05
        # ax2.set_ylim([ylim_min, ylim_max])
        ax2.set_xlim([t_set[0] - 5, t_set[2] + self.t_buffer + 5])

        fig.add_subplot(111, frameon=False)
        # hide tick and tick label of the big axis
        plt.tick_params(labelcolor='none', which='both',
                        top=False, bottom=False, left=False, right=False)
        plt.xlabel("time [min]")
        plt.ylabel(sweep_title + ' [' + units + ']')
        ax1.annotate('Experiment Overview', xy=(0.05, 0.95),
                     xycoords='axes fraction')
        line_names = ['Setpoint', "GC Sample"]
        if self.expt_type in ['comp_sweep', 'calibration']:
            line_names = self.gas_type + ["GC Sample"]

        ax2.legend(labels=line_names, loc='lower right')

        plt.tight_layout()

        run_time = t_set[-1]  # TODO break this out into separate func
        return (fig, ax1, ax2, run_time)

    def update_eqpt_list(self, eqpt_list):
        """
        Assign equipment objects as attributes of Experiment object.

        Takes eqpt_list as tuple in format
        (gc controller, laser controller, gas controller, heater)
        and assigns each component to experiment object
        updates sample rate by given value in gc_control and updates heater
        ramp rate by the rate specified in experiment object

        Parameters
        ----------
        eqpt_list: list[object]
            (:py:class:`~catalight.equipment.sri_gc.gc_control.GC_Connector`,
            :class:`~catalight.equipment.diode_laser.diode_control.Diode_Laser`,
            :class:`~catalight.equipment.alicat_MFC.gas_control.Gas_System`,
            :class:`~catalight.equipment.harrick_watlow.heater_control.Heater`)
        """
        # eqpt_list needs to be tuple
        self._gc_control = eqpt_list[0]
        self._laser_control = eqpt_list[1]
        self._gas_control = eqpt_list[2]
        self._heater = eqpt_list[3]

        if self.sample_rate >= self._gc_control.min_sample_rate:
            # try to set sample rate to that defined by expt
            self._gc_control.sample_rate = self.sample_rate
        else:
            # If that failed, _gc_control.sample_rate will be set to min
            # defined by ctrl file. This line makes sure gc and expt match
            self._sample_rate = self._gc_control.sample_rate
        self._heater.ramp_rate = self.heat_rate

    def set_initial_conditions(self):
        """Set initial conditions for experiment.

        Uses the first element of each attributes list to define initial
        conditions. Will also check that the temperature is not more than
        10 degrees C above the first setpoint.
        The order is:
        1. Set temperature
        2. Give 1 minute time warning for laser
        3. Set initial laser power
        4. Set gas type
        5. Set gas flows
        6. Wait 2 minutes
        7. Print gas flows
        8. Update gc sample set size
        9. Update date, time, and update log
        """
        unit = self.expt_list['Units'][0]
        self._heater.ramp(self.temp[0], temp_units=unit)
        starting_temp = self._heater.read_temp()
        starting_sp = self._heater.read_setpoint()
        print('Starting Temp = ' + str(starting_temp) + ' C')
        while starting_temp > (starting_sp + 10):
            print('Reactor is hotter than starting setpoint. Cooling...')
            time.sleep(120)
            starting_temp = self._heater.read_temp()

        if self.power[0] > 0:
            self._laser_control.time_warning(1)
            # Wait for a minute before turning on laser for safety
            time.sleep(60)

        self._laser_control.set_power(self.power[0])
        self._gas_control.set_gasses(self.gas_type)
        self._gas_control.set_flows(self.gas_comp[0], self.tot_flow[0])
        time.sleep(120)  # Wait for gas to steady out
        self._gas_control.print_flows()
        self._gc_control.sample_set_size = self.sample_set_size

        # Update experiment start time and date, save to log.
        self.update_date()
        self._start_time = time.time()  # Log beginning of expt
        # TODO Its a little silly to do all this instead of define a higher
        # level path as a instance attr in the first place.
        expt_path = os.path.dirname(self.data_path)
        self.update_expt_log(expt_path)

    def run_experiment(self):
        """
        Directs connected equipment to run experiment based on attributes.

        Most critical method of the class/package. This method directs the
        equipment to actually carry out the experiment based on the assigned
        attribute values for the object instance. This method currently works
        by using a series of if-statements to determine the experiment type
        and take the corresponding actions.
        """
        print('Starting ' + self.expt_type + self.expt_name)
        self.set_initial_conditions()
        step_num = 1
        # Creates subfolders for each step of experiment
        for step in getattr(self, self._ind_var):

            # Compare Boolean
            units = (self.expt_list['Units']
                     [self.expt_list['Active Status']].to_string(index=False))

            # This sets path for data storage according to expt type
            # ------------------------------------------------------
            if self._ind_var == 'gas_comp':  # merge gas_comp into name
                print(step)
                print(self.gas_type)
                step_str = '_'.join(
                    [str(m) + n for m, n in zip(step, self.gas_type)])
                path = os.path.join(self.data_path,
                                    ('%i %s%s' % (step_num, step_str, units)))
            elif self.expt_type == 'stability_test':
                path = os.path.join(self.data_path, '1 Stability_Test')
            else:
                path = os.path.join(self.data_path,
                                    ('%i %d%s' % (step_num, step, units)))
            step_num += 1

            # This chooses the run type and sets condition accordingly
            # --------------------------------------------------------
            # Python 3.10+ implements match/case which could replace these.
            if self.expt_type == 'temp_sweep':
                self._heater.ramp(step, temp_units=self.expt_list['Units'][0])
            elif self.expt_type == 'power_sweep':
                self._laser_control.set_power(step)
            elif self.expt_type in ['comp_sweep', 'calibration']:
                self._gas_control.set_flows(step, self.tot_flow[0])
                self._gas_control.print_flows()
            elif self.expt_type == 'flow_sweep':
                self._gas_control.set_flows(self.gas_comp[0], step)
                print(self.gas_type)
                print(np.array(self.gas_comp) * step)
                self._gas_control.print_flows()
            elif self.expt_type == 'stability_test':
                pass
                # Stability Test conditions set in initial conditions

            # This segment times when to start GC and prints status
            # -----------------------------------------------------
            print('Waiting for steady state: '
                  + time.strftime("%H:%M:%S", time.localtime()))
            t1 = time.time()
            while self._gc_control.is_running():
                time.sleep(10)  # Don't update ctrl file while running
            self._gc_control.update_ctrl_file(path)
            t2 = time.time()
            t_passed = round(t2 - t1)  # GC can take a while to respond
            for i in range(int(self.t_steady_state * 60 - t_passed)):
                time.sleep(1)  # Break sleep in bits so keyboard interrupt works

            print('Starting Collection: '
                  + time.strftime("%H:%M:%S", time.localtime()))
            print('Starting Temp = ', self._heater.read_temp(), ' C')
            self._gc_control.peaksimple.SetRunning(1, True)
            # t_collect ends on last gc pull
            t_collect = self.sample_rate * (self.sample_set_size - 1) * 60

            for i in range(int(t_collect)):
                time.sleep(1)

            print('Finished Collecting: '
                  + time.strftime("%H:%M:%S", time.localtime()))
            time.sleep(self.t_buffer * 60)
            print('Ending = ', self._heater.read_temp(), ' C')

            print('Step Finished: '
                  + time.strftime("%H:%M:%S", time.localtime()))

        print('Finished ' + self.expt_type + self.expt_name)


if __name__ == "__main__":
    # This is just a demo which runs when you run this class file as the main
    # script as of 20221026 this demo will not run because an extra mfc was
    # added to setup
    plt.close('all')
    # main_fol = ("C:\\Users\\brile\\Documents\\Temp Files\\"
    #             "20210524_8%AgPdMix_1wt%__200C_24.8mg\\PostReduction")
    main_fol = 'C:\\Users\\brile\\Documents\\Temp Files'
    Expt1 = Experiment()
    Expt1.expt_type = 'temp_sweep'
    Expt1.temp = list(np.arange(30, 150, 10) + 273)
    Expt1.gas_comp = [[0.01, 1 - 0.06, 0.05]]
    Expt1.tot_flow = [50]
    Expt1.sample_name = '20211221_fakesample'
    Expt1.plot_sweep()
    print('finished expt1')
    # Expt1.create_dirs(main_fol)

    Expt2 = Experiment()
    Expt2.expt_type = 'power_sweep'
    Expt2.power = list(np.arange(30, 150, 10))
    Expt2.gas_comp = [[0.01, 1 - 0.06, 0.05]]
    Expt2.tot_flow = [50]
    Expt2.plot_sweep()
    print('finished expt 2')
    # Expt2.create_dirs(main_fol)

    Expt3 = Experiment()
    Expt3.expt_type = 'flow_sweep'
    Expt3.temp = [273]
    Expt3.gas_comp = [[0.01, 1 - 0.06, 0.05]]
    Expt3.tot_flow = list(np.arange(10, 60, 10))
    Expt3.sample_name = '20211221_fakesample'
    _, _, _, t = Expt3.plot_sweep()
    print('finished expt3')

    Expt4 = Experiment()
    Expt4.expt_type = 'comp_sweep'
    Expt4.temp = [273]
    P_c2h2 = np.arange(0.5, 3.1, 0.5) / 100
    P_h2 = P_c2h2 * 5
    P_Ar = 1 - P_c2h2 - P_h2
    Expt4.gas_comp = np.stack([P_c2h2, P_Ar, P_h2], axis=1).tolist()
    Expt4.tot_flow = [50]
    Expt4.sample_name = '20211221_fakesample'
    Expt4.plot_sweep()
    print('finished expt4')

    Expt5 = Experiment()
    Expt5.expt_type = 'calibration'
    Expt5.gas_type = ['CalGas', 'Ar', 'H2']
    Expt5.temp = [273]
    P_CalGas = np.array([100, 50, 10]) / 100  # pretend one 1000ppm gas
    P_H2 = P_CalGas * 0
    P_Ar = 1 - P_CalGas - P_H2
    Expt5.gas_comp = np.stack([P_CalGas, P_Ar, P_H2], axis=1).tolist()
    Expt5.tot_flow = [50]
    Expt5.sample_name = '20211221_fakesample'
    Expt5.plot_sweep()
    # Expt5.create_dirs(main_fol)
    print('finished expt5')

    Expt6 = Experiment()
    Expt6.expt_type = 'stability_test'
    Expt6.gas_type = ['c2h2', 'Ar', 'H2']
    Expt6.temp = [100]
    P_c2h2 = 1 / 100
    P_h2 = P_c2h2 * 5
    P_Ar = 1 - P_c2h2 - P_h2
    Expt6.gas_comp = [[P_c2h2, P_Ar, P_h2]]
    Expt6.tot_flow = [50]
    Expt6.sample_name = '20210524_8%AgPdMix_1wt%_25mg'
    Expt6._date = '20211208'
    Expt6.plot_sweep()
    # Expt6.create_dirs(main_fol)
    print('finished Expt6')



"""
Contains the Heater class which connect to Watlow heating system.

Also contains convert_temp helper function used within class.

Created on Tue Feb  8 14:06:48 2022.
@author: Briley Bourgeois
"""

import os
import time

import numpy as np
import pandas as pd
from pywatlow.watlow import Watlow


def convert_temp(old_unit, new_unit, temp):
    """
    Converts temp in old_unit to temp in new_unit

    Parameters
    ----------
    old_unit : str
        Units to switch from (C, K, or F)
    new_unit : str
        Units to switch to (C, K, or F)
    temp : float or int
        Temperature in current units.

    Returns
    -------
    float
        Current temperature in new units.

    """
    # On call converts entry to key, value performs math on current temp.
    conversion = {'c_to_f': temp * 1.8 + 32.0,
                  'f_to_c': (temp - 32.0) / 1.8,
                  'k_to_f': (temp - 273) * 1.8 + 32.0,
                  'f_to_k': (temp - 32.0) / 1.8 + 273,
                  'c_to_k': temp + 273,
                  'k_to_c': temp - 273}
    return conversion[old_unit.lower() + '_to_' + new_unit.lower()]


class Heater:
    """
    Connect with and controls Watlow Heater.

    Allows ramping, reading set point and current temperature, and running test
    on heater, printing output
    """

    max_temp = 450
    """int or float: Max temp in C; 450C reduces lifetime 900C is real max.
    To change, use Heater.max_temp = new_value"""

    def __init__(self):
        """Connect to watlow, print current state."""
        self.controller = Watlow(port='COM5', address=1)
        self.is_busy = False  #: Used to block access from other threads
        print('Heater Initializing...')
        print('Current temperature = ' + str(self.read_temp()) + ' C')
        print('Current setpoint = ' + str(self.read_setpoint()) + ' C')
        self.ramp_rate = 15  #: C/min used for ramp method
        # TODO put check on heat rate

    def read_temp(self, temp_units='C'):
        """
        Read current controller temp in defined units (default = C).

        Parameters
        ----------
        temp_units : str, optional
            (C, K, or F) units to return temp in. The default is 'C'.

        Returns
        -------
        float
            Current temperature in requested units.

        """
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True
        temp = self.controller.read()['data']
        self.is_busy = False
        if temp_units.upper() != 'F':
            temp = convert_temp('F', temp_units, temp)
        return round(temp, 3)

    def read_setpoint(self, temp_units='C'):
        """
        Read current controller setpoint in defined units (default = C).

        Parameters
        ----------
        temp_units : str, optional
            (C, K, or F) units to return temp in. The default is 'C'.

        Returns
        -------
        float
            Current setpoint in requested units, rounded to 3 digits.

        """
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True
        setpoint = self.controller.readSetpoint()['data']
        self.is_busy = False
        if temp_units.upper() != 'F':
            setpoint = convert_temp('F', temp_units, setpoint)
        return round(setpoint, 3)

    def shut_down(self):
        """Set heater to 0 F."""
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True
        self.controller.write(0)
        self.is_busy = False

    def disconnect(self):
        """Run shut_down() and then close connection."""
        self.shut_down()
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True
        self.controller.close()
        self.is_busy = False

    def ramp(self, T2, T1=None, temp_units='C', record=False):
        """
        Ramp the heater from T1 to T2 at ramp rate defined by instance attr.

        If record=True, records the time, setpoint, and temp. Plots outcome.

        Parameters
        ----------
        T2 : float
            Target setpoint
        T1 : float
            Starting setpoint Uses the last setpoint if None. None is default.
        temp_units : str
            C, K, or F. The default is 'C'.
        record : bool
            If True, will record setpoints during ramp, plot outcome,
            and return readout dataframe and plot. Mainly used for testing
            heater performance.

        Returns
        -------
        pandas.DataFrame, when record=True
            pandas.DataFrame with columns ['time', 'set point', 'temperature']
            recorded during the experiment.
        matplotlib.pyplot.figure, when record=True
            figure handle for ramp rate plot
        matplotlib.pyplot.axis, when record=True
            axis handle for ramp rate plot
        """
        if T1 is None:
            T1 = self.read_setpoint(temp_units)

        # Switch units to C
        if temp_units.upper() == 'F':
            T2, T1 = [convert_temp('F', 'C', temp) for temp in [T2, T1]]
        elif temp_units.upper() == 'K':
            T2, T1 = [convert_temp('K', 'C', temp) for temp in [T2, T1]]
        elif temp_units.upper() != 'C':
            raise ValueError('Temp Unit not recognized')

        if T2 > Heater.max_temp:
            print('Desired setpoint is above maximum safe working temp!!!')
            print('Resetting T2 to heater max instead...')
            T2 = Heater.max_temp

        refresh_rate = 20  # 1/min
        ramp_time = (T2 - T1) / self.ramp_rate  # min
        setpoints = np.linspace(T1, T2, abs(int(ramp_time * refresh_rate)))
        read_out = []
        for temp in setpoints:
            temp = convert_temp('C', 'F', temp)  # Change units to F
            if record:
                read_out.append([time.time(),
                                 self.read_setpoint(),
                                 self.read_temp()])
            while self.is_busy:
                time.sleep(0)
            self.is_busy = True
            self.controller.write(temp)  # write to controller
            self.is_busy = False
            time.sleep(60 / refresh_rate)  # wait

        print('Soak Temp = ' + str(self.read_temp()))

        if record:  # Record and return output
            read_out = pd.DataFrame(read_out, columns=['time', 'set point',
                                                       'temperature'])
            read_out['time'] = (read_out['time'] - read_out['time'][0]) / 60
            ax = read_out.plot(x='time')
            ax.set_xlabel('time [min]')
            ax.set_ylabel('Temperature [$\degree$C]')
            ax.set_title('Ramp rate = ' + str(self.ramp_rate))
            return (read_out, ax.get_figure(), ax)
        else:
            return

    def test_heater_performance(self, savepath, rates, T_max, T_min=30):
        """
        Loops through provided ramp rates, ramps heater, save outcomes.

        Sets heater to 20C, gives 5 min to steady out. Loops through ramp rates
        while recording the outcome (Heater.ramp(record=True)). Saves results.
        Always lets heater cool to < 30C between runs.

        Parameters
        ----------
        savepath : str
            Full path to directory in which to save data.
        rates : list[int or float]
            (deg C/min) List of heater ramp rates to test e.g. [5, 10, 15, 20]
        T_max : int or float
            (deg C) Maximum temperature setpoint to use for testing
        T_min : int or float, optional
            (deg C) Starting setpoint to use for testing. The default is 30.

        Returns
        -------
        None.

        """
        heater.ramp(20)  # Start from 'off'
        time.sleep(300)  # Allow steady state for 5 min
        for rate in rates:
            # Set new rate, ramp, turn off
            heater.ramp_rate = rate
            read_out, fig, ax = heater.ramp(T_max, T_min, record=True)
            heater.shut_down()

            # Allow temperature to return to < 30 C
            while heater.read_temp() > 30:
                time.sleep(60)

            # Save results
            read_out.to_csv(os.path.join(savepath,
                                         str(rate) + '_heater_read_out.csv'))
            fig.savefig(os.path.join(savepath, str(rate) + '_heater_test.svg'),
                        format="svg")


if __name__ == "__main__":
    # Example Usage
    heater = Heater()
    heater.ramp(20)
    print(heater.read_setpoint())
    print(heater.read_temp())



"""
Tools specifically related to plotting data.

This module was broken out from analysis.tools for clarity. Helper functions
and command specifically related to plotting data of any type should be placed
here.
Created on Mon Feb  6 11:46:49 2023.
@author: Briley Bourgeois
"""
import os
import pickle
import re

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import catalight.analysis.tools as analysis_tools


def set_plot_style(figsize=(6.5, 4.5)):
    """
    Update default plot styles. Varies font size with figure size.

    Parameters
    ----------
    figsize : tuple, optional
        Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
        figsize suggestions:
        1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
        1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);
    """
    if figsize[0] < 6:
        fontsize = [11, 14]
    elif figsize[0] > 7:
        fontsize = [16, 20]
    else:
        fontsize = [14, 18]

    # Some Plot Defaults
    plt.rcParams['axes.linewidth'] = 2
    plt.rcParams['lines.linewidth'] = 1.5
    plt.rcParams['xtick.major.size'] = 6
    plt.rcParams['xtick.major.width'] = 1.5
    plt.rcParams['ytick.major.size'] = 6
    plt.rcParams['ytick.major.width'] = 1.5
    plt.rcParams['figure.figsize'] = figsize
    plt.rcParams['font.size'] = fontsize[0]
    plt.rcParams['axes.labelsize'] = fontsize[1]
    # plt.rcParams['text.usetex'] = False
    plt.rcParams['svg.fonttype'] = 'none'


def plot_expt_summary(expt, calDF, reactant, target_molecule, mole_bal='c',
                      figsize=(6.5, 4.5), savedata='True', switch_to_hours=2):
    """
    Call 'typical' plotting functions and save output to results path.

    Plots and saves "run_num", "ppm", and "X and S" for given experiment.
    run_num: ppm concentration as a function of run number
    ppm: average ppm concentration w/ error for each experiment condition
    X and S: Plots the average conversion and selectivity as a function of
    experimental condition

    If savedata=True, will save figures as both .svg and .pickle

    Parameters
    ----------
    expt : Experiment
        Experiment object which has data and has been analyzed
        (i.e. has avg_conc.csv etc.)
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end]
    reactant : str
        String identity of reactant molecule to track. Must match what exists
        in the calibration file exactly.
    target_molecule : str
        String identity of the target to use when calculating selectivity. Must
        match what exists in the calibration file exactly.
    mole_bal : str, optional
        Code will perform a mole balance for the element provided.
        The default is 'c'. (i.e. carbon balance)
    figsize : tuple, optional
         Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
         figsize suggestions:
         1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
         1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);
    savedata : bool, optional
        Indicates whether or not to save data. The default is 'True'.
    switch_to_hours : float, optional
        Time in hours when the output should switch units to
        hours instead of minutes. The default is 2.

    Returns
    -------
    tuple(:class:`~matplotlib.axes._axes.Axes`,
    :class:`~matplotlib.axes._axes.Axes`,
    :class:`~matplotlib.axes._axes.Axes`) :
        (ax1, ax2, ax3) Axis handles for "run_num",  "ppm", and "X and S" plots

    """
    # Plotting
    ###########################################################################
    print('Plotting...')
    plt.close('all')

    set_plot_style(figsize)
    fig1, ax1 = plot_run_num(expt, calDF, switch_to_hours)
    fig2, ax2 = plot_ppm(expt, calDF, mole_bal, switch_to_hours)
    fig3, ax3 = plot_X_and_S(expt, reactant, target_molecule)

    # Figure Export
    if savedata:
        fig1_path = os.path.join(expt.results_path,
                                 str(figsize[0]) + 'w_run_num_plot')
        fig2_path = os.path.join(expt.results_path,
                                 str(figsize[0]) + 'w_avg_conc_plot')
        fig3_path = os.path.join(expt.results_path,
                                 str(figsize[0]) + 'w_Conv_Sel_plot')

        fig1.savefig(fig1_path + '.svg', format="svg")
        fig2.savefig(fig2_path + '.svg', format="svg")
        fig3.savefig(fig3_path + '.svg', format="svg")

        pickle.dump(fig1, open(fig1_path + '.pickle', 'wb'))
        pickle.dump(fig2, open(fig2_path + '.pickle', 'wb'))
        pickle.dump(fig3, open(fig3_path + '.pickle', 'wb'))

    return (ax1, ax2, ax3)


def plot_run_num(expt, calDF, switch_to_hours=2):
    """
    Plot the individual concentrations for each molecule vs. run number.

    Parameters
    ----------
    expt : Experiment
        Experiment object which has data and has been analyzed
        (i.e. has avg_conc.csv etc.)
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end]
    switch_to_hours : float, optional
        Time in hours when the output should switch units to
        hours instead of minutes. The default is 2.

    Returns
    -------
    tuple(matplotlib.figure.Figure, matplotlib.axes._axes.Axes) :
        Figure and Axis handle for "run_num" plot
    """
    # Initialize run num plot
    fig, ax = plt.subplots()
    concentrations, avg, std = analysis_tools.load_results(expt)
    # Calculations:
    calchemIDs = calDF.index  # get chem IDs from calibration files
    time_passed, time_unit = analysis_tools.get_timepassed(concentrations,
                                                           switch_to_hours,
                                                           expt)

    # use regex to determine number of carbons (or other) in molecule name
    for chemical in calchemIDs:
        chem_num = calchemIDs.get_loc(chemical) + 1  # index 0 is timestamp
        # Concentrations for individual chemical
        ind_concentrations = concentrations[:, chem_num, :]
        ind_concentrations = ind_concentrations[~np.isnan(ind_concentrations)]
        if sum(ind_concentrations) == 0:
            continue  # Skip chemicals with no values
        ax.plot(time_passed, ind_concentrations, 'o', label=chemical)

    ax.legend()
    plt.xlabel('Time [' + time_unit + ']')
    # TODO add second x axis as run number instead of time
    plt.ylabel('Conc [ppm]')
    plt.tight_layout()
    return (fig, ax)


def plot_ppm(expt, calDF, mole_bal='c', switch_to_hours=2):
    """
    Plot average concentration in ppm for each molecule vs reaction condition.

    Parameters
    ----------
    expt : Experiment
        Experiment object which has data and has been analyzed
        (i.e. has avg_conc.csv etc.)
    concentrations : numpy.ndarray
        3D matrix of concentrations for each molecule, gc collection,
        and condition
        [Condition x [Timestamps, ChemID] x run number
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end]
    mole_bal : str, optional
        Code will perform a mole balance for the element provided.
        The default is 'c'. (i.e. carbon balance)
    switch_to_hours : float, optional
        Time in hours when the output should switch units to
        hours instead of minutes. The default is 2.

    Returns
    -------
    tuple(matplotlib.figure.Figure, matplotlib.axes._axes.Axes) :
        Figure and Axis handle for "ppm" plot
    """
    # Initialize ppm vs ind_var plot
    fig, ax = plt.subplots()
    concentrations, avg, std = analysis_tools.load_results(expt)
    calchemIDs = calDF.index  # get chem IDs from calibration files

    units = (expt.expt_list['Units']
             [expt.expt_list['Active Status']].to_string(index=False))

    if expt.expt_type == 'stability_test':
        # TODO This could check unit if expt_list updates
        # to have time instead of temp in future.
        if np.max(avg.index) > switch_to_hours:
            avg.index = avg.index / 60  # convert minutes to hours
            units = 'hr'

    stoyk = pd.Series(0, index=calchemIDs)
    # use regex to determine number of carbons (or other) in molecule name
    for chemical in calchemIDs:
        # Use chem_num to access by index number instead of name.
        # chem_num = calchemIDs.get_loc(chemical) + 1  # index 0 is timestamp
        # read number after 'c' in each chem name
        count = re.findall(mole_bal + r"(\d+)", chemical, re.IGNORECASE)
        if not count:
            if re.findall(mole_bal, chemical, re.IGNORECASE):
                count = 1
            else:
                count = 0
        else:
            count = int(count[0])

        stoyk[chemical] = count

    mol_count = avg @ (np.array(stoyk, dtype=int))
    mol_count.columns = ['Total ' + mole_bal]
    mol_error = std @ (np.array(stoyk, dtype=int))

    # Plotting:
    # Don't plot molecules that don't show up
    undetected = avg.sum(axis=0) == 0
    if avg.to_numpy().sum() == 0:
        print('No Molecules Detected')
        return (fig, ax)

    avg.loc[:, ~undetected].plot(ax=ax, marker='o',
                                 yerr=std.loc[:, ~undetected])
    mol_count.plot(ax=ax, marker='o', yerr=mol_error)
    ax.set_xlabel(expt.ind_var + ' [' + units + ']')
    ax.set_ylabel('Conc (ppm)')
    # ax.set_xticklabels(avg.iloc[0:len(x_data)], rotation=90)
    plt.tight_layout()
    return (fig, ax)


def plot_X_and_S(expt, reactant, target_molecule):
    """
    Plot conversion and selectivity vs reaction condition with error.

    This method will call calculate_X_and_S() to generate the plotting values.

    Parameters
    ----------
    expt : Experiment
        Experiment object which has data and has been analyzed
        (i.e. has avg_conc.csv etc.)
    reactant : str
        String identity of reactant molecule to track. Must match what exists
        in the calibration file exactly.
    target_molecule : str
        String identity of the target to use when calculating selectivity. Must
        match what exists in the calibration file exactly.

    Returns
    -------
    tuple(matplotlib.figure.Figure, matplotlib.axes._axes.Axes) :
        Figure and Axis handle for "X and S" plot

    """
    # Initialize Conv and Selectivity plot
    fig, ax = plt.subplots()
    results = analysis_tools.calculate_X_and_S(expt, reactant, target_molecule)
    units = (expt.expt_list['Units']
             [expt.expt_list['Active Status']].to_string(index=False))
    X = results['Conversion']
    S = results['Selectivity']
    if 'X Error' in results.columns:
        X_err = results['X Error']
        S_err = results['S Error']
    else:  # Older data calculated one Error
        X_err = results['Error'] * results['Conversion']
        S_err = results['Error'] * results['Selectivity']
    X.plot(ax=ax, yerr=X_err, fmt='--ok')
    S.plot(ax=ax, yerr=S_err, fmt='--^r')
    ax.set_xlabel(expt.ind_var + ' [' + units + ']')
    ax.set_ylabel('Conv./Selec. [%]')
    plt.legend(['Conversion', 'Selectivity'])
    ax.set_ylim([0, 105])
    plt.tight_layout()
    return (fig, ax)


def multiplot_X_and_S(results_dict, figsize=(6.5, 4.5)):
    """
    Separately plot 1) conversion and 2) selectivity for multiple experiments.

    Produces two plots:
        (1) conversion vs reaction condition
        (2) selectivity vs reaction condition

    Each plot can have multiple experiments listed. The independent variable
    should be the same between all experiments provided.

    Parameters
    ----------
    results_dict : dict
        Dictionary of results (produced by calculate_X_and_S) in the form
        {data label: result}
        where data label is a string with the desired legend label and
        result is a pandas.DataFrame with column headers
        ['Conversion', 'Selectivity', 'Error']
        This function is most easily used in conjunction with DataExtractor
        such as in the run_plot_comparison script.
    figsize : tuple, optional
        Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
        figsize suggestions:
        1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
        1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);

    Returns
    -------
    tuple(tuple(matplotlib.figure.Figure, matplotlib.axes._axes.Axes)) :
        two tuples, each containing the (figure, axis) handles for the
        conversion and selectivity plots, respectively.

    """
    # Initialize Conv and Selectivity plots
    # Note: if desired, results_dict could instead be expt_dict and
    # calculate_X_and_S() called within this function
    figX, axX = plt.subplots()
    figS, axS = plt.subplots()
    xlabel = None
    for data_set in results_dict.items():
        data_label = data_set[0]
        results = data_set[1]
        X = results['Conversion']
        S = results['Selectivity']
        if 'X Error' in results.columns:
            X_err = results['X Error']
            S_err = results['S Error']
        else:
            X_err = results['Error'] * results['Conversion']
            S_err = results['Error'] * results['Selectivity']
        X.plot(ax=axX, yerr=X_err, fmt='--o', label=data_label)
        S.plot(ax=axS, yerr=S_err, fmt='--^', label=data_label)
        xlabel = results.index.name

    axX.set_ylabel('Conversion [%]')
    axX.set_xlabel(xlabel)
    axX.set_ylim([0, 105])
    plt.legend()
    figX.tight_layout()

    axS.set_ylabel('Selectivity [%]')
    axS.set_xlabel(xlabel)
    axS.set_ylim([0, 105])
    plt.legend()
    figS.tight_layout()
    figX.show()
    figS.show()

    return ((figX, axX), (figS, axS))


def multiplot_X_vs_S(results_dict, figsize=(6.5, 4.5)):
    """
    Produce a combo plot of selectivity as a function of conversion.

    This function can compare multiple experiments, and the independent
    variables do not need to match.

    Parameters
    ----------
    results_dict : dict
        Dictionary of results (produced by calculate_X_and_S) in the form
        {data label: result}
        where data label is a string with the desired legend label and
        result is a pandas.DataFrame with column headers
        ['Conversion', 'Selectivity', 'Error']
        This function is most easily used in conjunction with DataExtractor
        such as in the run_plot_comparison script.
    figsize : tuple, optional
        Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
        figsize suggestions:
        1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
        1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);

    Returns
    -------
    tuple(matplotlib.figure.Figure, matplotlib.axes._axes.Axes) :
        Figure and Axis handle for "X vs S" plot

    """
    print('Plotting...')
    plt.close('all')

    set_plot_style(figsize)
    # Initialize Conv and Selectivity plot
    fig, ax = plt.subplots()

    for data_set in results_dict.items():
        data_label = data_set[0]
        results = data_set[1]
        X = results['Conversion']
        S = results['Selectivity']
        X_err = results['X Error']
        S_err = results['S Error']
        # Defunct. Older data had single error output.
        # X_err = results['Error'] * results['Conversion']
        # S_err = results['Error'] * results['Selectivity']

        ax.errorbar(X, S, yerr=S_err, xerr=X_err, fmt='--o', label=data_label)

    ax.set_ylabel('Selectivity [%]')
    ax.set_xlabel('Conversion [%]')
    ax.set_xlim([0, 105])
    ax.set_ylim([0, 105])
    plt.legend()
    fig.tight_layout()
    fig.show()
    return (fig, ax)


def open_pickled_fig(fig_path):
    """
    Open a pickled figure for editing.

    Parameters
    ----------
    fig_path : str
        Path to .pickle file. Must be matplotlib.pyplot.figure

    Returns
    -------
    tuple(matplotlib.figure.Figure, matplotlib.axes._axes.Axes) :
        Figure and Axis handle for plot
    """
    plt.rcParams['svg.fonttype'] = 'none'
    fig = pickle.load(open(fig_path, 'rb'))
    ax = fig.get_axes()[0]
    fig.show()  # Show the figure, edit it, etc.!
    return (fig, ax)



"""
Graphical User Interface for scanning through chromatograms within a directory.

@authors: Claire Carlin, Briley Bourgeois
"""
import os
import sys

import pyqtgraph as pg
import numpy as np
from catalight.analysis.gcdata import GCData
import catalight.analysis.tools as analysis_tools
from PyQt5.QtWidgets import (QMainWindow, QWidget, QVBoxLayout, QGridLayout,
                             QCheckBox, QComboBox, QListWidget, QFileDialog,
                             QApplication, QLineEdit, QLabel)

os.environ['QT_MAC_WANTS_LAYER'] = '1'
# function which updates graph based on highlighted list item


class MainWindow(QMainWindow):
    """Subclass QMainWindow to customize your application's main window."""

    def __init__(self):
        super().__init__()
        self.setWindowTitle("Use Arrow Keys to Cycle Chromatograms")

        # Create widgets
        self.layout = QVBoxLayout()
        self.widget = QWidget()  # Main widget
        self.listWidget = QListWidget()
        self.graphWidget = pg.PlotWidget()

        # Create options
        self.options_layout = QGridLayout()
        self.bc_box = QCheckBox("baseline correction?")
        self.int_box = QCheckBox("Plot integration bounds?")
        self.target_entry = QComboBox()
        self.target_entry.addItems(['FID', 'TCD'])
        self.suffix_entry = QLineEdit('.asc')
        self.target_label = QLabel('Data Type:')
        self.suffix_label = QLabel('File Ending:')

        # Assemble options to layout
        self.options_layout.addWidget(self.bc_box, 0, 0)
        self.options_layout.addWidget(self.int_box, 1, 0)
        self.options_layout.addWidget(self.target_label, 0, 1)
        self.options_layout.addWidget(self.target_entry, 1, 1)
        self.options_layout.addWidget(self.suffix_label, 0, 2)
        self.options_layout.addWidget(self.suffix_entry, 1, 2)

        # Add widgets to layout
        self.layout.addWidget(self.listWidget)
        self.layout.addLayout(self.options_layout)
        self.layout.addWidget(self.graphWidget)

        # Add layout to window
        self.widget.setLayout(self.layout)
        self.setCentralWidget(self.widget)

        # Connect Signals/Slots
        # Update which plot shows up to be whichever is highlighted
        self.listWidget.currentItemChanged.connect(self.update_plot)
        self.target_entry.currentTextChanged.connect(self.get_files)
        self.suffix_entry.editingFinished.connect(self.get_files)
        self.bc_box.clicked.connect(self.update_plot)
        self.int_box.clicked.connect(self.update_plot)

        # Get main directory containing data files
        self.folderpath = QFileDialog \
            .getExistingDirectory(None, 'Select folder containing data')
        self.set_graph_style()
        self.get_files()
        self.init_plot()

    def get_files(self):
        """List user's desired files and add to listWidget."""
        # Populate list with all datafiles in selected folder based on options
        filematch_args = ([self.folderpath],
                          self.target_entry.currentText(),
                          self.suffix_entry.text())
        self.filelist = analysis_tools.list_matching_files(*filematch_args)
        if not self.filelist:
            print('No matching files found in folder')
        self.listWidget.clear()  # Make sure widget is empty
        self.listWidget.addItems(self.filelist)

    def init_plot(self):
        """Pull first data file and add to plot."""
        # Make the first file in the list appear as the plot upon opening
        # Get data from filepath
        path = self.listWidget.item(0).text()
        data = GCData(path, basecorrect=False)
        datacorr = GCData(path, basecorrect=True)

        # Plot pulled data
        self.data_line = self.graphWidget.plot(data.time, data.signal,
                                               pen=self.pen1, name="raw")
        self.data_line_corr = self.graphWidget.plot(datacorr.time,
                                                    datacorr.signal,
                                                    pen=self.pen2,
                                                    name="corrected")
        self.data_line_corr.clear()
        self.graphWidget.setTitle('Run 0:', color="k", size="18pt")

    def set_graph_style(self):
        """Update graph appearance."""
        # graph styling things
        self.graphWidget.addLegend(labelTextColor="k", labelTextSize="14pt")
        self.graphWidget.setBackground('w')
        styles = {'color': 'r', 'font-size': '18px'}
        self.graphWidget.setLabel('left', 'Counts', **styles)
        self.graphWidget.setLabel('bottom', 'Time (min)', **styles)
        self.graphWidget.getAxis('left').setTextPen('k')
        self.graphWidget.getAxis('bottom').setTextPen('k')
        pen = pg.mkPen(color=(0, 0, 0), width=2)
        self.graphWidget.plotItem.getAxis('left').setPen(pen)
        self.graphWidget.plotItem.getAxis('bottom').setPen(pen)
        self.graphWidget.plotItem.getAxis('right').setPen(pen)
        self.graphWidget.plotItem.getAxis('top').setPen(pen)

        # Pens for later plotting
        self.pen1 = pg.mkPen(color=(0, 102, 255), width=2)
        self.pen2 = pg.mkPen(color=(0, 153, 51), width=1)

    def update_plot(self):
        """Update plot with current listWidget item, if present."""
        if not self.listWidget.currentItem():
            print('No Items')
            self.data_line.clear()
            self.data_line_corr.clear()
            self.graphWidget.setTitle('')
            return

        path = self.listWidget.currentItem().text()
        data = GCData(path, basecorrect=False)
        datacorr = GCData(path, basecorrect=True)
        num = self.filelist.index(path)

        # Update title of graph
        self.graphWidget.setTitle("run " + str(num) + ":",
                                  color="k", size="18pt")

        # Update data, check if baseline correction is needed
        self.data_line.setData(data.time, data.signal,
                               pen=self.pen1, name="raw")
        if self.int_box.isChecked():
            data.integration_inds()
            [left_idx, right_idx] = (np.rint(data.lind).astype('int'),
                                     np.rint(data.rind).astype('int'))
            self.graphWidget.plot(data.time[data.apex_ind],
                                  data.signal[data.apex_ind],
                                  'bo', label='apex')
            self.graphWidget.plot(data.time[left_idx],
                                  data.signal[left_idx],
                                  'go', label='left')
            self.graphWidget.plot(data.time[right_idx],
                                  data.signal[right_idx],
                                  'ro', label='right')
        if self.bc_box.isChecked():
            self.data_line_corr.setData(datacorr.time, datacorr.signal,
                                        pen=self.pen2, name="corrected")
        else:
            self.data_line_corr.clear()


if __name__ == "__main__":

    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec_())



"""
Generate conversion and selectivity plots for multiple experiments using GUI.

Created on Fri Jun 24 17:28:48 2022
@author: Briley Bourgeois
"""
import sys

import matplotlib.pyplot as plt
from PyQt5.QtWidgets import QApplication
from catalight import analysis
from catalight.analysis.user_inputs import (DataExtractor,
                                            PlotOptionsDialog,
                                            PlotOptionList)


def get_user_inputs(starting_dir=None):
    """
    Take user input using GUI for running multiplot comparison.

    Parameters
    ----------
    starting_dir : str, optional
        Path to initialize file dialog in. The default is None.

    Returns
    -------
    file_list : list of str
        List of file paths to experiment folders then used to initiate
        experiment objects and calculate conversion and selectivity.
    data_labels : list of str
        List of data labels used for generating plot legends.
    response_dict : dict
        Dictionary of user plot options.
        Dict Keys depend on options proved to PlotOptionsDialog.
        {'reactant': str, 'target_molecule': str,
        'XandS': bool, 'XvsS': bool, 'figsize': tuple}
    """
    app = QApplication(sys.argv)  # noqa

    # Prompt user to select multiple experiment directories and label them
    data_dialog = DataExtractor(starting_dir)
    if data_dialog.exec_() == DataExtractor.Accepted:
        file_list, data_labels = data_dialog.get_output()

    # Edit Options specifically for initial analysis dialog
    include_dict = {'reactant': True, 'target_molecule': True,
                    'plot_XandS': True, 'plot_XvsS': True, 'figsize': True}
    options = PlotOptionList()  # Create default gui options list
    options.change_includes(include_dict)  # Modify gui components
    options_dialog = PlotOptionsDialog(options)  # Build dialog w/ options
    if options_dialog.exec_() == PlotOptionsDialog.Accepted:
        response_dict = options.values_todict()

    return (file_list, data_labels, response_dict)


def main(file_list, data_labels, reactant, target_molecule,
         plot_XandS, plot_XvsS, figsize=(6.5, 4.5)):
    """
    Plot multi-experiment comparison.

    Parameters
    ----------
    file_list : list of str
        List of file paths to experiment folders then used to initiate
        experiment objects and calculate conversion and selectivity.
    data_labels : list of str
        List of data labels used for generating plot legends.
    reactant : str
        String identity of reactant molecule to track. Must match what
        exists in the calibration file exactly.
    target_molecule : str
        String identity of the target_molecule to use when calculating
        selectivity. Must match what exists in the calibration file exactly.
    plot_XandS : bool
        True plots comparison plots of conversion and selectivity individually
        as a function of the independent variable. Ind. Var. should be the same
        between all provided values.
    plot_XvsS : bool
        True plots plot selectivity as a function of conversion for all
        provided experiments on one plot.
    figsize : tuple, optional
         Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
         figsize suggestions:
         1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
         1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);

    Returns
    -------
    tuple(tuple, tuple):
        ((figX, axX), (figS, axS)) if plot_XandS == True,
        returns fig and ax handles for two plots
    matplotlib.figure.Figure :
        if plot_XvsS == True, returns fig handle for selectivity vs
        conversion plot
    matplotlib.axes._axes.Axes :
        if plot_XvsS == True, returns ax handle for selectivity vs
        conversion plot

    """
    results_dict = analysis.tools.build_results_dict(file_list, data_labels,
                                                     reactant, target_molecule)

    if plot_XandS:
        plot_results = analysis.plotting.multiplot_X_and_S(results_dict,
                                                           figsize)
        ((figX, axX), (figS, axS)) = plot_results
        plt.show()
        return ((figX, axX), (figS, axS))

    elif plot_XvsS:
        fig, ax = analysis.plotting.multiplot_X_vs_S(results_dict, figsize)
        plt.show()
        return fig, ax
    print('Test if this line is executed')


if __name__ == "__main__":
    plt.close('all')
    file_list, data_labels, response_dict = get_user_inputs()
    main(file_list, data_labels, **response_dict)



"""
Contains heater resources and code.

Contains resources (manuals, calibration files, etc.) and code for connecting
with equipment

Created on Mon Jan 16 16:29:11 2023
@author: Briley Bourgeois
"""



"""
Port connection tester module returns the comport name/address of MFCs.

Utilizes the is_connected method of the alicat module's FlowController and
FlowMeter classes. Usual serial module to list all comports on computer, then
sweeps through every alphabetic letter in alphabet to checking for MFC
connections. Each comport loop breaks if a MFC is found. Prints summary at end.

Created on Wed Jan 19 15:32:48 2022
@author: Briley Bourgeois
"""
from string import ascii_uppercase as alphabet

import serial.tools.list_ports
from alicat import FlowController, FlowMeter

if __name__ == '__main__':
    comports = list(serial.tools.list_ports.comports())
    positive_cases = []

    print('Beginning search: This process will take several minutes')
    for port in comports:
        port_name = port.name
        for address_name in alphabet:
            test_case_str = (port_name + ' ' + address_name)
            if FlowController.is_connected(port_name, address=address_name):
                print(test_case_str + ' is flow controller')
                positive_cases.append(test_case_str + ' is flow controller\n')
                break
            elif FlowMeter.is_connected(port_name, address=address_name):
                print(test_case_str + ' is flow meter')
                positive_cases.append(test_case_str + ' is flow meter\n')
                break
            else:
                print(test_case_str + ' is not MFC')

    print('Search concluded. Results:')
    print(*positive_cases)



"""
Gas chromatograph data and associated processing methods.

GCData is meant to represent a gas chormatograph data file and associated
data processing techniques. Individual data files are operated on
by this class, then processed data for an experiment is
compiled, analyzed, and plotted elsewhere.
"""
import datetime as dt
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.ndimage as nd
import scipy.signal as scisig


class GCData:
    """
    A class for processing/analyzing gas chromatograph data from an ASCII file.

    Attributes
    ----------
    filepath : (str)
        filepath where data is stored (.ASC file)
    timestamp : (float)
        time since the epoch in seconds
    rawdata : pandas.DataFrame
        contains time and signal values from GC run
    time : numpy.ndarray
        np array version of time axis from rawdata (in minutes)
    signal : numpy.ndarray
        np array version of time axis from rawdata.
        If basecorrect=True, it will be baseline
        corrected using baseline_correct()
    apex_ind : numppy.ndarray
        indices of all peaks identified by apex_inds()
    numpeaks : int
        length of apex_ind
    lind : numpy.ndarray
        indices of leftmost bound for integration for each peak identified
    rind : numpy.ndarray
        indices of rightmost bound for integration for each peak identified
    """

    def __init__(self, filepath, basecorrect=False):
        """
        Initialize the class with the attributes filename and data.

        (which has been read from ASCII and is a pandas dataframe)

        Parameters
        ----------
        filepath : str
            full path to the acsii file
        basecorrect : bool
            set basecorrect to True if you want correction, default is false
        """
        self.filepath = filepath
        self.timestamp, self.rawdata = self.getrawdata()
        self.time = np.asarray(self.rawdata['Time'])  # in minutes
        if basecorrect:
            self.signal = self.baseline_correction()
        else:
            self.signal = np.asarray(self.rawdata['Signal'])
        self.apex_ind = self.apex_inds()
        self.numpeaks = len(self.apex_ind)
        self.lind, self.rind = self.integration_inds()

    def getrawdata(self):
        """
        Read data from ASCII, returns pandas dataframe.

        Uses Matthias Richter's example code (translated from Matlab)
        to read GC .ASC files

        Returns
        -------
        tuple(float, pandas.DataFrame)
            (timestamp, [Elution Time, GC Signal])
            First element is timestamp in time since epoch. Second element is
            pandas.DataFrame containing 'Time' (min) and 'Signal' (arb units)
        """
        with open(self.filepath, 'r') as f:
            # Skip first 18 lines
            for i in range(18):
                next(f)
            # Line 19, date
            [month, day, year] = [int(i.strip()) for i in
                                  f.readline().split("=")[1].split('-')]
            # Line 20, time
            [hr, minute, second] = [int(i.strip()) for i in
                                    f.readline().split("=")[1].split(':')]
            timestamp = dt.datetime(year, month, day,
                                    hr, minute, second).timestamp()
            # Line 21, sampling rate
            rate = int(f.readline().split('=')[1][0])
            # Line 22 # of data points
            size = int(f.readline().split("=")[1])
            # Skip next 3 lines
            for i in range(3):
                next(f)

            y = []
            for line in f:
                if line.strip() == '':  # Empty strings in between entries
                    next
                elif "IPOINT" in line:  # Ignore IPOINT numbers at the bottom
                    next
                else:
                    value = int(line.strip().split(',')[0])
                    y.append(value / 1000)  # Convert mV to V

            x = np.linspace(0, 1 / rate / 60 * (size - 1), num=size).tolist()
            raw_data = pd.DataFrame({'Time': x, 'Signal': y})
            GC_data = (timestamp, raw_data)
        return GC_data

    # Processing functions
    ###########################################################################

    def baseline_correction(self):
        """
        Get self.signal, output signal with background subtraction.

        Uses tophat filter, based on PyMassSpec/pyms/TopHat.py

        Returns
        -------
        numpy.ndarray
            GC signal with baseline corrected
        """
        self.signal = np.asarray(self.rawdata['Signal'])
        # default struct element as frac of tot num points
        struct_elm_frac = 0.1
        struct_pts = int(round(self.signal.size * struct_elm_frac))
        str_el = np.repeat([1], struct_pts)
        line = nd.generate_binary_structure(rank=1, connectivity=9)  # noqa
        # structure = line
        signal_basesub = nd.white_tophat(input=self.signal, footprint=str_el)
        return signal_basesub

    def apex_inds(self):
        """
        Use scipy.signal.find_peaks to find peaks in signal.

        Returns
        -------
        numpy.ndarray
            All peak locations (as integer indices NOT times)
        """
        apex_ind, _ = scisig.find_peaks(self.signal, prominence=4)
        return apex_ind

    def integration_inds(self):
        """
        Find bounds of integration for apexes.

        Edge values determined by self._half_index_search.
        Output in the same (time) order as the apex_ind array with the location
        (index) of the left and right points for integration

        Returns
        -------
        numpy.ndarray of int
            Left bounds of integration.
        numpy.ndrray of int
            Right bounds of integration.

        """
        lind = np.zeros(len(self.apex_ind))
        rind = np.zeros(len(self.apex_ind))
        k = 0

        for apex in self.apex_ind:
            # select all of the data just past the apex
            lhs = self.signal[:apex + 1]
            # reverse list so that working forward works toward the left
            flhs = np.flip(lhs)
            lind[k] = apex - self._half_index_search(flhs)
            rhs = self.signal[apex - 1:]
            rind[k] = apex + self._half_index_search(rhs)
            k += 1

        return lind.astype(int), rind.astype(int)

    @staticmethod
    def _half_index_search(dat, tol=0.5):
        """
        Support function for integration_inds to search for integration bounds.

        Determined where:
        1. the change in signal is less than 0.5%
            of the previous signal point (with averaging); or
        2. the added intensity starts increasing
            (i.e. when the ion is common to co-eluting compounds)

        adapted from pyMS function peak_sum_area
        by Andrew Isaac and Sean O'Callaghan
        https://github.com/ma-bio21/pyms/blob/master/pyms/Peak/Function.py

        Parameters
        ----------
        tol: float, optional
            Tolerance in percent. Summing stops when change in sum is
            less than tol percent of the current area.

        Returns
        -------
        int
            index of the integration bound
        """
        # convert from percent, not sure why it should also be halved
        tol = tol / 200.0
        # number of points to sum new area across
        # (increasing value increases smoothing)
        wide = 10
        sig = dat[0]
        limit = len(dat)
        index = 1
        # edge is average value of num of pts defined by wide (NOT an index)
        edge = float(sum(dat[0:wide])) / wide
        delta = sig - edge
        # make sure old_edge starts as larger than the current edge
        old_edge = 2 * edge
        # look for change is large, edge going down, limit not hit
        while abs(delta) > sig * tol and edge < old_edge and index < limit:
            old_edge = edge
            sig = dat[index]
            edge = float(sum(dat[index:index + wide])) / wide
            delta = sig - edge
            index += 1

        index -= 1
        return index

    def integrate_peak(self):
        """
        Find the area under the peak using a trapezoidal method.

        Calculate for each peak identified using bounds from integration_inds.
        Integrates peak trapezoidal in units of seconds*signal intensity.

        Returns
        -------
        numpy.int32
            counts rounded using np.around()
        """
        counts = np.zeros(self.numpeaks)
        for i in range(0, self.numpeaks):
            counts[i] = np.trapz(self.signal[self.lind[i]:self.rind[i]],
                                 x=60 * self.time[self.lind[i]:self.rind[i]])
            if counts[i] == 0:
                counts[i] = 1
        return np.around(counts)

    def get_concentrations(self, calDF):
        """
        Return a Pandas series of chemical concentrations.

        Output has the same order as the calibration file chem IDs.
        Prints warning if zero molecules are detected or if peaks are found
        outside of those listed in calDF.

        Parameters
        ----------
        calDF: pandas.DataFrame
            Calibration values by chemical ID

        Returns
        -------
        pandas.Series
            Concentrations for each peak in apex_ind, given in ppm
        Notes
        -----
            Unknown peaks could be added to calibration dataframe for reference
        """
        # TODO Unknown peaks could be added
        # to calibration dataframe for reference
        self.integration_inds()
        # Creates empty series where index are ChemIDs from Cal file
        conc = pd.Series(0.0, index=['timestamp', *calDF.index.to_list()])
        conc['timestamp'] = self.timestamp
        counts = self.integrate_peak()
        UnknownPeaks = 0
        for i in range(len(self.apex_ind)):
            # Determine peak time based on index
            peak_time = self.time[self.apex_ind[i]]
            # determine if peak falls within range for any calibration data set
            match = calDF[(calDF["start"] < peak_time)
                          & (peak_time < calDF["end"])].index
            if not match.empty:  # if empty, there are no matches
                chemID = match[0]  # if not empty, name match chemical:chemID
                # Convert counts to ppm append to conc list for the ChemID
                conc[chemID] = (self.convert_to_ppm(calDF, counts[i], chemID))
            else:
                UnknownPeaks += 1
                # TODO I can take the ind_x and append this
                # unknown peak to the calibration df

        if UnknownPeaks > 1:  # Theres always a peak from back flush right now
            print('Warning: %5d Unknown peaks detected' % (UnknownPeaks))
            print(self.filepath)
        if (conc == 0).all():  # Checks if all concentrations are zero
            print('Warning: Zero Molecules Detected')

        return conc

    # Plotting functions
    ###########################################################################

    def plot_integration(self):
        """
        Plot the chromatogram with the peaks and indices indicated.

        Notes
        -----
            The plotting style update here can be integrated with data_analysis
            module.
        """
        # TODO The plotting style update here can be integrated
        # with data_analysis module.
        plt.rcParams.update({'font.size': 14})
        plt.rcParams['axes.linewidth'] = 2
        plt.gca().tick_params(which='both', width=1.5, length=6)
        plt.gca().tick_params(which='minor', width=1.5, length=3)

        plt.figure
        self.integration_inds()
        [left_idx, right_idx] = (np.rint(self.lind).astype('int'),
                                 np.rint(self.rind).astype('int'))
        # Plot signal
        plt.plot(self.time, self.signal, linewidth=2.5)
        # Plot Derivative
        # plt.plot(time, 10*np.gradient(signal))
        # Plot peak and bounds
        plt.plot(self.time[self.apex_ind], self.signal[self.apex_ind],
                 'bo', label='apex')
        plt.plot(self.time[left_idx], self.signal[left_idx],
                 'go', label='left')
        plt.plot(self.time[right_idx], self.signal[right_idx],
                 'ro', label='right')

        plt.xlabel('Retention (min)', fontsize=18)
        plt.ylabel('Signal (a.u.)', fontsize=18)
        plt.legend()
        plt.tight_layout()
        plt.show()

    # Helper functions
    ###########################################################################

    def get_run_number(self):
        """
        Determine run number from filename.

        Returns
        -------
        int
            run number extracted from filename
        """
        filename = os.path.basename(self.filepath)
        parts = filename.split('_')
        run_number = int(parts[-1][-2:].lstrip("0"))
        return run_number

    def convert_to_ppm(self, calDF, counts, chemID):
        """
        Convert integrated raw counts into ppm based on calibration data.

        Parameters
        ----------
        calDF: pandas.DataFrame
            Calibration values by chemical ID
        counts: float
            Raw integrated counts determined by integrate_peaks()
        chemID: str
            Contains name of chemical from calDF to be converted to ppm

        Returns
        -------
        float
            Counts for a single peak converted into ppm based on calDF
        """
        # Pull Cal data for given chemical
        [m, b] = (calDF.loc[chemID, ['slope', 'intercept']])
        # Simple calibration equation. Needs update if calibration isn't linear
        y = m * counts + b
        return y  # Counts in ppm



"""
Use of Ophir COM object.

Works with python 3.5.1 & 2.7.11
Uses pywin32
This is a copy and pasted demo from the guide
"""
import time
import traceback

import win32com.client
import win32gui

try:
    OphirCOM = win32com.client.Dispatch("OphirLMMeasurement.CoLMMeasurement")
    # Stop & Close all devices
    OphirCOM.StopAllStreams()
    OphirCOM.CloseAll()
    # Scan for connected Devices
    DeviceList = OphirCOM.ScanUSB()
    print(DeviceList)
    for Device in DeviceList:   	# if any device is connected
        DeviceHandle = OphirCOM.OpenUSBDevice(Device)  # open first device
        exists = OphirCOM.IsSensorExists(DeviceHandle, 0)
        if exists:
            print('\n----------Data for S/N {0} ---------------'.format(Device))

            # An Example for Range control. first get the ranges
            ranges = OphirCOM.GetRanges(DeviceHandle, 0)
            print(ranges)
            # change range at your will
            if ranges[0] > 0:
                newRange = ranges[0] - 1
            else:
                newRange = ranges[0] + 1
                # set new range
                OphirCOM.SetRange(DeviceHandle, 0, newRange)

            # An Example for data retrieving
            OphirCOM.StartStream(DeviceHandle, 0)		# start measuring
            for i in range(10):
                time.sleep(.2)				# wait a little for data
                data = OphirCOM.GetData(DeviceHandle, 0)
            if len(data[0]) > 0:		# if any data available, print the first one from the batch
                print('Reading = {0}, TimeStamp = {1}, Status = {2} '
                      .format(data[0][0], data[1][0], data[2][0]))

        else:
            print('\nNo Sensor attached to {0} !!!'.format(Device))

except OSError as err:
    print("OS error: {0}".format(err))
except Exception as e:
    print(e)
    traceback.print_exc()

win32gui.MessageBox(0, 'finished', '', 0)
# Stop & Close all devices
OphirCOM.StopAllStreams()
OphirCOM.CloseAll()
# Release the object
OphirCOM = None



"""
Created on Thu Oct 14 17:30:17 2021

This script is an example of how to import the PeaksimpleConnector class into
Python using the python.NET package
@author: Briley Bourgeois
"""

import os

import clr  # python.NET is imported with the name clr (Common Language Runtime)

dir_path = os.path.dirname(os.path.realpath(__file__))
assemblydir = os.path.join(dir_path, 'PeaksimpleClient', 'PeaksimpleConnector.dll')

clr.AddReference(assemblydir)  # Add the assembly to python.NET

# Now that the Assembly has been added to python.NET,
# it can be imported like a normal module
import Peaksimple  # Import the assembly namespace, which has a different name

Connector = Peaksimple.PeaksimpleConnector()  # This class has all the functions

ctrl_file = 'C:/Program Files/Added Programs/PeakSimple version 4.89Win10/DEFAULT2.CON'

def initialize_connection():
    # Connect to running instance of peaksimple using class method
    Connector.Connect()



"""
Access Newport power meter through Use of Ophir COM object.

You should need to PMManager installed on your computer to use this.
"""
import time

import win32com.client


class NewportMeter:
    """NewportMeter."""

    def __init__(self):
        # Ophir COM Object is main tool for this class
        self.OphirCOM = win32com.client.Dispatch("OphirLMMeasurement.CoLMMeasurement")

        # Stop & Close all devices
        self.OphirCOM.StopAllStreams()
        self.OphirCOM.CloseAll()

        # Scan for connected Devices
        DeviceList = self.OphirCOM.ScanUSB()
        if len(DeviceList) > 1:
            raise ValueError('Multiple Devices Found')
        elif len(DeviceList) == 0:
            raise ValueError('No Devices Found')
        else:
            self.Device = DeviceList[0]

        # Open device
        self.DeviceHandle = self.OphirCOM.OpenUSBDevice(self.Device)
        exists = self.OphirCOM.IsSensorExists(self.DeviceHandle, 0)
        if exists:
            print('Connected to S/N {0}'.format(self.Device))
        else:
            print('No Sensor attached to {0} !!!'.format(self.Device))

    def change_wavelength(self, wavelength):
        """Change to desired wavelength, modifies slot 4 if wavelength DNE."""
        wavelengths = self.OphirCOM.GetWavelengths(self.DeviceHandle, 0)[1]
        wavelength = str(wavelength)
        if wavelength not in wavelengths:
            self.OphirCOM.ModifyWavelength(self.DeviceHandle, 0, 4, wavelength)
        idx = wavelengths.index(wavelength)
        self.OphirCOM.SetWavelength(self.DeviceHandle, 0, idx)

    def read(self, averaging_time=0.5):
        """
         Read average powermeter output in mW over averaging_time.

        Parameters
        ----------
        averaging_time : float or int, optional
            Time to collect dataset in seconds. The default is 0.5.

        Returns
        -------
        timestamp : float
            Time since epoch from time.time() of recording
        reading : float
            Power reading in mW
        """
        self.OphirCOM.SetRange(self.DeviceHandle, 0, 0)  # set range to auto
        self.OphirCOM.StartStream(self.DeviceHandle, 0)  # start measuring
        time.sleep(averaging_time)  # collect data during pause
        data = self.OphirCOM.GetData(self.DeviceHandle, 0)
        # average the readout, convert units to mW
        reading = sum(list(data[0])) / len(data[0]) * 1000
        timestamp = time.time()  # get time
        self.OphirCOM.StopAllStreams()  # stop measuring
        return(timestamp, reading)

    def shut_down(self):
        """Stop & Close all devices."""
        self.OphirCOM.StopAllStreams()
        self.OphirCOM.CloseAll()
        # Release the object
        self.OphirCOM = None


if __name__ == "__main__":
    power_meter = NewportMeter()
    print(power_meter.read())



"""
Module used to calibrate diode laser and save calibration data.

User does not need to update the file, except to change the wavelength of the
diode laser being used. Works for Newport power meter and Thorlabs diode driver
powered by MCC DAQ.
Created on Tue Feb 22 23:15:23 2022
@author: Briley Bourgeois
"""
import numpy as np
import pandas as pd
from catalight.equipment.diode_laser.diode_control import Diode_Laser
from catalight.equipment.power_meter.power_meter_ctrl import NewportMeter


def main(current_range=(150, 800, 35), wavelength=450, tolerance=10):
    """
    Runs a calibration on the connected Diode_Laser.

    The current is varied based on the supplied current_range and the power
    is measured using the attached NewportMeter. The sampling time is the
    default defined in the NewportMeter class (0.5 sec/sample). Data is
    collected until 20 samples and the desired tolerance factor is met.
    Calibration results are saved locally to be accessed when initializing the
    Diode_Laser class. A plot showing the fit is produced. Only linear fits
    are supported at the moment (2023/02/09).

    Hardware Configuration: The laser should be on, connected to the DAQ board,
    and the current turned to 0 on the current controller. The power meter
    needs to be placed in the beam path at the location in which the user would
    like to calibrate the output (i.e. at the sample)

    Parameters
    ----------
    current_range : tuple, optional
        (I_min, I_max, step) tuple sent to range(). Used to control the testing
        range. Current cycled from I_min to I_max in steps.
        The default is (150, 800, 35).
    wavelength : int or float, optional
        Wavelength of diode. The default is 450.
    tolerance : int or float, optional
        Tolerance for each data point in percent. Sampling will continue until
        20 points have been collected and the standard deviation is less than
        the requested tolerance. Default sampling time is 0.5 seconds/point.
        The default is 10.

    Returns
    -------
    fig : matplotlib.pyplot.figure
        Figure handle for calibration plot.
    ax : matplotlib.pyplot.axis
        Axis handle for calibration plot.

    """

    diode = Diode_Laser()
    power_meter = NewportMeter()
    power_meter.change_wavelength(wavelength)

    data = pd.DataFrame(columns=['Power', 'error'])
    print('Running Calibration...')

    for current in range(current_range):
        diode.set_current(current)
        error = 100
        power_readings = np.array([-1000])  # Set unreal number
        while error > tolerance:
            power_readings = np.append(power_readings, power_meter.read()[1])
            if len(power_readings) > 20:
                power_readings = np.delete(power_readings, 0)
                std = np.std(power_readings)
                avg = np.average(power_readings)
                error = std / avg * 100

        print('Power Reading = %.2f +/- %.2f\n' % (avg, std))
        data.loc[current] = [avg, std]
    x_data, y_data, y_err = (data.index, data['Power'], data['error'])
    ax = data.plot(y=['Power'], ylabel='Power (mW)',
                   xlabel='Current (mA)', yerr=y_err)
    try:
        p, V = np.polyfit(x_data, y_data, 1, cov=True, w=1 / y_err)
        m, b, err_m, err_b = (*p, np.sqrt(V[0][0]), np.sqrt(V[1][1]))
        label = '\n'.join(["m: %4.2f +/- %4.2f" % (m, err_m),
                           "b: %4.2f +/- %4.2f" % (b, err_b)])
    except (np.linalg.LinAlgError):
        p = np.polyfit(x_data, y_data, 1)
        m, b = (p[0], p[1])
        label = '\n'.join(["m: %4.2f" % m,
                           "b: %4.2f" % b])
    x_fit = np.linspace(0, max(x_data), 100)
    ax.plot(x_fit, (p[0] * x_fit + p[1]), '--r')
    diode.update_calibration(p[0], p[1])
    diode.shut_down()
    fig = ax.get_figure()
    fig.savefig('diode_laser/calibration_plot.svg', format="svg")
    print(label)
    return fig, ax


if __name__ == "__main__":
    main()



"""
Executable script prompting the user to select experiments and enter new xdata.

Created on Sat Mar 19 15:33:01 2022
@author: Briley Bourgeois
"""
import sys

import pandas as pd
import matplotlib.pyplot as plt
import catalight.analysis.tools as analysis_tools
from PyQt5.QtWidgets import (QApplication, QFileDialog)
from catalight.analysis.user_inputs import (DataExtractor,
                                               PlotOptionsDialog,
                                               PlotOptionList)


def get_user_inputs(starting_dir=None, cal_folder=None):
    """
    Get user inputs for changing x axis data.

    Parameters
    ----------
    starting_dir : str, optional
        Path to initialize file dialog in. The default is None.
    cal_folder : str, optional
        Path containing calibration files. The default in None.

    Returns
    -------
    file_list : list
        list of paths to expt_logs
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end]
    response_dict : dict
        Dictionary of user plot options.
        Dict Keys depend on options proved to PlotOptionsDialog.
        {'xdata': list, 'units': str, 'reactant': str,
        y'target_molecule': str, 'mole_bal': str, 'figsize': tuple,
        'savedata': bool, 'switch_to_hours': float}

    """
    # Prompt user to select calibration file
    app = QApplication(sys.argv)
    cal_file = QFileDialog.getOpenFileName(None, 'Select Calibration File',
                                           cal_folder, "csv file (*.csv)")[0]
    calDF = pd.read_csv(cal_file, delimiter=',', index_col='Chem ID')

    # Prompt user to select multiple experiment directories. Label unused.
    data_dialog = DataExtractor(starting_dir)
    if data_dialog.exec_() == DataExtractor.Accepted:
        file_list, data_labels = data_dialog.get_output()

    # Edit Options specifically for initial analysis dialog
    include_dict = {'xdata': True, 'units': True, 'reactant': True,
                    'target_molecule': True, 'mole_bal': True, 'figsize': True,
                    'savedata': True, 'switch_to_hours': True}
    options = PlotOptionList()  # Create default gui options list
    options.change_includes(include_dict)  # Modify gui components
    options_dialog = PlotOptionsDialog(options)  # Build dialog w/ options
    if options_dialog.exec_() == PlotOptionsDialog.Accepted:
        response_dict = options.values_todict()

    return file_list, calDF, response_dict


def main(expt_paths, calDF, new_x, units, reactant, target_molecule,
         mole_bal='c', figsize=(6.5, 4.5), savedata=False, switch_to_hours=2):
    """
    Change the x axis data for previously plotted experiments.

    Parameters
    ----------
    expt_paths : list of str
        List of full paths to expt_log.txt files.
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end]
    new_x : list
        Array of new data to be used inplace of old x axis
    units : str
        new units for new x axis data
    reactant : str
        String identity of reactant molecule to track. Must match what exists
        in the calibration file exactly.
    target_molecule : str
        String identity of the target to use when calculating selectivity. Must
        match what exists in the calibration file exactly.
    mole_bal : str, optional
        Code will perform a mole balance for the element provided.
        The default is 'c'. (i.e. carbon balance)
    figsize : tuple, optional
         Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
         figsize suggestions:
         1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
         1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);
    savedata : bool, optional
        Indicates whether or not to save data. The default is 'True'.
    switch_to_hours : float, optional
        Time in hours when the output should switch units to
        hours instead of minutes. The default is 2.

    Returns
    -------
    None.

    """
    options = (calDF, reactant, target_molecule, mole_bal,
               figsize, savedata, switch_to_hours)

    expts = analysis_tools.list_expt_obj(expt_paths)
    for expt in expts:
        # Must have previously analyzed data
        (results, avg, std) = analysis_tools.load_results(expt)
        # Redefine x axis and update units
        avg.index = new_x
        std.index = new_x
        expt.expt_list['Units'][2] = units
        # Replot data using new axes data
        (ax1, ax2, ax3) = analysis_tools.plot_expt_summary(expt, *options)


if __name__ == "__main__":
    plt.close('all')
    file_list, calDF, response_dict = get_user_inputs()
    main(file_list, calDF, **response_dict)



"""
Module containing Gas System class used for coordinating multiple MFCs

Created on Sun Feb 13 20:56:51 2022.
@author: Briley Bourgeois
"""
import os
import time

import numpy as np
import pandas as pd
from alicat import FlowController, FlowMeter


class Gas_System:
    """MFC control."""

    factory_gasses = ['C2H2', 'Air', 'Ar', 'i-C4H10', 'n-C4H10', 'CO2', 'CO',
                      'D2', 'C2H6', 'C2H4', 'He', 'H2', 'Kr', 'CH4', 'Ne',
                      'N2', 'N2O', 'O2', 'C3H8', 'SF6', 'Xe']
    #: Factory gas list saved to Alicat MFCs

    def __init__(self):
        """
        Connect MFCs based on COM port and address. User needs to update.

        The user needs to update the address and COM ports for each mfc
        based on their specific setup. This process can be assisted by using
        the "connection_tester.py" file
        """
        self.mfc_A = FlowController(port='COM6', address='A')
        self.mfc_B = FlowController(port='COM9', address='B')
        self.mfc_C = FlowController(port='COM8', address='C')
        self.mfc_D = FlowController(port='COM10', address='D')
        self.mfc_E = FlowMeter(port='COM11', address='E')
        self.is_busy = False

    def set_gasses(self, gas_list):
        """
        Update gas type to MFCs based on input list.

        Parameters
        ----------
        gas_list : list of str
            [gasA, gasB, gasC, gasD] Each must be within factory gasses.

        Returns
        -------
        None.

        """
        # Custom mixes cannot be indexed by name in alicat package
        # mix 237 is designated as the calibration gas slot in this package
        # if calgas is given in gas_list, assign via mix number instead of name
        gas_list_copy = gas_list.copy()  # Don't edit global!!
        for i in range(len(gas_list_copy)):
            if gas_list_copy[i].lower() == 'calgas':
                gas_list_copy[i] = 237

        while self.is_busy:
            time.sleep(0)

        self.is_busy = True
        self.mfc_A.set_gas(gas_list_copy[0])
        self.mfc_B.set_gas(gas_list_copy[1])
        self.mfc_C.set_gas(gas_list_copy[2])
        self.mfc_D.set_gas(gas_list_copy[3])
        self.is_busy = False

    def set_flows(self, comp_list, tot_flow):
        """
        Set flow rates based as fraction of tot_flow from comp_list.

        Sets the flow rate of all mfc based on a desired total flow
        and the desired gas composition. Also call set_gasE
        TODO: add limit for tot_flow for each MFC

        Parameters
        ----------
        comp_list : list of float
            list of gas fraction for mfc [a, b, c, d]. Must sum to 1 or 100
        tot_flow : float
            Total flow to send.

        Raises
        ------
        AttributeError
            if gas comp doesn't sum to 1 or 100

        Returns
        -------
        None.

        """
        comp_list = self.check_comp_total(comp_list)
        while self.is_busy:
            time.sleep(0)

        self.is_busy = True
        self.mfc_A.set_flow_rate(float(comp_list[0] * tot_flow))
        self.mfc_B.set_flow_rate(float(comp_list[1] * tot_flow))
        self.mfc_C.set_flow_rate(float(comp_list[2] * tot_flow))
        self.mfc_D.set_flow_rate(float(comp_list[3] * tot_flow))
        self.is_busy = False

        self.set_gasE(comp_list)

    def set_gasE(self, comp_list):
        """
        Set gas E as a mixture of input gasses based on comp. and types.

        Parameters
        ----------
        comp_list : list of float
            list of gas fraction for mfc [a, b, c, d]. Must sum to 1 or 100

        Returns
        -------
        None.

        """
        comp_list = self.check_comp_total(comp_list)

        while self.is_busy:
            time.sleep(0)

        self.is_busy = True
        gas_list = []
        for mfc in [self.mfc_A, self.mfc_B, self.mfc_C, self.mfc_D]:
            gas_name = mfc.get()['gas']
            if gas_name.lower() == 'calgas':
                gas_list.append('Ar')  # Can't handle custom mixes
            else:
                gas_list.append(gas_name)
        # convert to percents, make dict, drop zero values
        percents = np.array(comp_list, dtype=float) * 100
        gas_series = pd.Series(percents, gas_list)
        gas_series = gas_series.groupby(level=0).sum()  # sums duplicates
        gas_dict = gas_series.to_dict()
        gas_dict = {x:y for x,y in gas_dict.items() if y != 0}

        # Uses create_mix method to write to gas slot 236,
        # first custom gas slot on MFC
        if len(gas_dict)>1: # if more than 1 gas, creates mix
            self.mfc_E.create_mix(mix_no=236, name='output',
                                  gases=gas_dict)
            self.mfc_E.set_gas(236)
        else:  # If only one gas, sets that as output
            self.mfc_E.set_gas(list(gas_dict)[0])
        self.is_busy = False

    def check_comp_total(self, comp_list):
        """
        Takes in gas composition list, checks sum, returns as fractions.

        Parameters
        ----------
        comp_list : list of float or int
            Composition list in either percents or fractions.

        Raises
        ------
        AttributeError
            Composition list doesn't sum to 1 or 100

        Returns
        -------
        comp_list : list of float
            Updated composition list as fractions

        """

        if sum(comp_list) == 100:  # convert % to fraction
            comp_list[:] = [x / 100 for x in comp_list]

        if (sum(comp_list) != 1) and (sum(comp_list) != 0):
            raise AttributeError('Gas comp. must be list of list == 1')

        return comp_list

    def print_flows(self):
        """Print mass flow rates and gas type for each MFC to console."""
        while self.is_busy:
            time.sleep(0)

        self.is_busy = True
        print('MFC A = ' + str(self.mfc_A.get()['mass_flow'])
              + self.mfc_A.get()['gas'])
        print('MFC B = ' + str(self.mfc_B.get()['mass_flow'])
              + self.mfc_B.get()['gas'])
        print('MFC C = ' + str(self.mfc_C.get()['mass_flow'])
              + self.mfc_C.get()['gas'])
        print('MFC D = ' + str(self.mfc_D.get()['mass_flow'])
              + self.mfc_D.get()['gas'])
        print('MFC E = ' + str(self.mfc_E.get()['mass_flow'])
              + self.mfc_E.get()['gas'])
        self.is_busy = False

    def print_details(self):
        """
        Run mfc.get() for each mfc, printing full status details.

        Returns
        -------
        None.

        """
        while self.is_busy:
            time.sleep(0)

        self.is_busy = True
        print(self.mfc_A.get())
        print(self.mfc_B.get())
        print(self.mfc_C.get())
        print(self.mfc_D.get())
        print(self.mfc_E.get())
        self.is_busy = False

    def read_flows(self):
        """
        Return full details of each mfc condition arranged in a dict.

        Returns
        -------
        flow_dict : dict of dict
            {mfc: mfc.get()}
        """
        while self.is_busy:
            time.sleep(0)

        self.is_busy = True
        flow_dict = {'mfc_A': self.mfc_A.get(),
                     'mfc_B': self.mfc_B.get(),
                     'mfc_C': self.mfc_C.get(),
                     'mfc_D': self.mfc_D.get(),
                     'mfc_E': self.mfc_E.get()}
        self.is_busy = False
        return (flow_dict)

    def shut_down(self):
        """Set MFC with Ar or N2 running to 1 sccm and others to 0."""
        while self.is_busy:
            time.sleep(0)

        self.is_busy = True
        mfc_list = [self.mfc_A, self.mfc_B, self.mfc_C, self.mfc_D]
        for mfc in mfc_list:
            if mfc.get()['gas'] in ['Ar', 'N2']:
                mfc.set_flow_rate(1.0)
            else:
                mfc.set_flow_rate(0.0)
        self.is_busy = False

    def disconnect(self):
        """Call Gas_System.shut_down then disconnect from MFCs."""
        self.shut_down()
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True
        self.mfc_A.close()
        self.mfc_B.close()
        self.mfc_C.close()
        self.mfc_D.close()
        self.mfc_E.close()
        self.is_busy = False
        del self

    def set_calibration_gas(self, mfc, calDF, fill_gas='Ar'):
        """
        Sets a custom gas mixture for the mfc of choice.

        Typically for calibration gas. Creates a gas mixture based on the
        largest 5 components calgas. Sets new mixture to slot 237 as name
        'CalGas'. This function uses the standard calDF format utilized
        elsewhere in this code. Please consult the alicat gas composer website
        for the official list of possible gasses.

        Parameters
        ----------
        mfc : alicat.FlowController | alicat.FlowMeter
            Mass flow controller or meter to update with calgas
        calDF : pandas.DataFrame
            Formatted DataFrame containing gc calibration data.
            Specific to control file used!
            Format [ChemID, slope, intercept, start, end]
        fill_gas : str
            Inert gas dilutant for calibration gas tank. The default is 'Ar'.
        """
        percents = calDF['ppm'] / 10000
        percents.index = percents.index.map(lambda x: x.split('_')[0])
        percents = percents[~percents.index.duplicated()]
        percents = percents.sort_values(ascending=False)
        new_idx = pd.DataFrame([False] * len(percents), index=percents.index)
        for chemical in percents.index:
            new_idx.loc[chemical] = chemical in Gas_System.factory_gasses
        percents = percents[new_idx[0]]
        percents = percents[0:4]  # TODO Handle if gas list short
        percents = percents.round(2)  # High precision breaks FlowController()
        percents[fill_gas] = 100 - percents.sum()
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True
        print(percents.to_dict())
        mfc.create_mix(mix_no=237, name='CalGas', gases=percents.to_dict())
        self.is_busy = False

    def test_pressure(self, savepath, flows, num_samples=5):
        """
        Ramp flow rate, measure pressure, plot results, save.

        Ramp flow rate from 5-50 sccm in steps of 5, measure pressure every
        minute for num_samples times. Waits 1 min before first collection.
        Plots results on yyplot w/ time on x-axis, and pressure and flow on y.

        Parameters
        ----------
        savepath : str
            Path to folder to save the results in.
        flows : list[int or float]
            Flow rate setpoints to sweep through when testing pressure build up
        num_samples : int, optional
            Number of samples to collect at each flow rate. Default is 5.
        """
        print('Testing Pressure Build-up...')
        output = pd.DataFrame(columns=['time', 'setpoint',
                                       'flow rate', 'pressure'])
        mfc_list = [self.mfc_A, self.mfc_B, self.mfc_C, self.mfc_D]
        sample_num = 0
        # Last MFC registered as inert gas gets used as tester.
        for mfc in mfc_list:
            if mfc.get()['gas'] in ['Ar', 'N2']:
                test_mfc = mfc
            else:
                mfc.set_flow_rate(0.0)

        self.mfc_E.set_gas(test_mfc.get()['gas'])
        print('Starting Conditions:')
        self.print_flows()
        start_time = time.time()
        # Loop through flow rate, record reading, save to output
        for setpoint in flows:
            test_mfc.set_flow_rate(setpoint)
            for sample in range(0, num_samples):
                time.sleep(60)
                pressure = test_mfc.get()['pressure']
                flow_rate = self.mfc_E.get()['mass_flow']
                reading = [(time.time() - start_time) / 60,
                           setpoint, flow_rate, pressure]
                print('time: %4.1f (min) setpoint: %4.2f (sccm) '
                      'flow rate: %4.2f (sccm) pressure: %4.2f (psia)'
                      % tuple(reading))
                output.loc[sample_num] = reading
                sample_num += 1
        # Plot Results
        ax1 = output.plot(x='time', y='pressure',
                          ylabel='Pressure (psia)', style='--ok')
        ax2 = ax1.twinx()
        ax2.spines['right'].set_position(('axes', 1.0))
        output.plot(ax=ax2, x='time',
                    y=['setpoint', 'flow rate'], ylabel='Flow Rate (sccm)')
        fig = ax1.get_figure()
        # Save Results
        fig.savefig(os.path.join(savepath, 'flow_test.svg'), format='svg')
        output.to_csv(os.path.join(savepath, 'flow_test.csv'))


if __name__ == "__main__":
    gas_controller = Gas_System()
    comp_list = [0, 0, 0, 1]
    tot_flow = 1
    gas_controller.set_flows(comp_list, tot_flow)



'''
Create empty calibration experiment data structure.

This script was initially used to generate a "dummy" calibration experiment
using data collected before calibration experiments were automated. Use this
script if you have calibration data that is not organized in the appropriate
data structure. This will create expt log file and data layout. then copy paste
calibration data into appropriate folders.
'''
import numpy as np
from catalight.equipment.experiment_control import Experiment


def main(main_fol, filename, dilution, tot_flow=1):
    """
    Build an empty set of folders to put calibration data into.

    The main point of this module is to replicate a composition sweep
    experiment for users that will be performing a calibration using multiple
    sets of calibration gasses. This package generally assumes that calibration
    is performed using one calibration gas tank diluted with MFCs. If this is
    not your situation, this module may be useful.

    Parameters
    ----------
    main_fol : str
        Directory to save folders in.
    filename : str
        Folder name you'd like to use when saving 'calibration expt'
    dilution : list of float or int
        Percent amount you diluted the original calibration gas
        For example, you have 1000 ppm, 500 ppm, and 100 ppm calibration gasses
        This would mimic diluting a 1000 ppm gas by 100%, 50%, 10%
        So the list you provide would be [100, 50, 10]
    tot_flow : float or int, optional
        Enter if you'd like to specify/record flow rate used. The default is 1.

    Returns
    -------
    None.

    """

    expt = Experiment()
    expt.expt_type = 'calibration'
    expt.gas_type = ['CalGas', 'Ar', 'Ar', 'Ar']
    P_CalGas = np.array(dilution) / 100   # pretend one 1000ppm gas
    P_Ar = P_CalGas * 0
    expt.gas_comp = np.stack([P_CalGas, P_Ar, P_Ar], axis=1).tolist()
    expt.tot_flow = tot_flow
    expt.sample_name = filename
    expt.create_dirs(main_fol)


if __name__ == "__main__":
    # Example usage:
    main_fol = "/Users/ccarlin/Documents/calibration"
    tot_flow = 50
    filename = '20200202_calibration'
    # Percent amount you diluted the original calibration gas
    # For example, you have 1000 ppm, 500 ppm, and 100 ppm calibration gasses
    # This would mimic diluting a 1000 ppm gas by 100%, 50%, 10%
    dilution = [100, 50, 10]
    main(main_fol, filename, dilution, tot_flow)



"""
Executable file that asks user for inputs and plot GC data on one plot.

Opens 3 UIs requesting, data directory, target .asc file w/ labels and plotting
options. Plots .asc file data on a single figure.

Created on Fri Jan  6 17:10:25 2023.
@author: Briley Bourgeois
"""
import sys

import matplotlib.pyplot as plt
from PyQt5.QtWidgets import QApplication
import catalight.analysis.plotting as plot_tools
from catalight.analysis.gcdata import GCData
from catalight.analysis.user_inputs import (DataExtractor,
                                            PlotOptionList,
                                            PlotOptionsDialog)


def get_user_inputs(starting_dir=None):
    """
    Prompt user to enter select directory and choose .asc files from it.

    Opens 3 UI screens. Prompts user for directory, then displays .asc files
    within that directory and asks user to click the ones to plot and provide
    a data label, finally opens a UI to ask for plotting options
    (baseline correction and figure size)

    Parameters
    ----------
    starting_dir : str, optional
        Directory to start file dialog in. The default is None.

    Returns
    -------
    file_list : list of str
        List of files selected by user.
    data_labels : list of str
        List of data labels provided by user for plot legend.
    response_dict : dict
        Dictionary of user plot options.
        Dict Keys depend on options proved to PlotOptionsDialog.
        {'basecorrect': bool, 'figsize': tuple}

    """
    app = QApplication(sys.argv)  # noqa
    # Prompt user to select .asc files and label them
    data_dialog = DataExtractor(starting_dir, '', '.asc', data_depth=0)
    if data_dialog.exec_() == DataExtractor.Accepted:
        file_list, data_labels = data_dialog.get_output()

    # Open UI and request plot options from user
    include_dict = {'basecorrect': True, 'figsize': True}
    options = PlotOptionList()  # Create default gui options list
    options.change_includes(include_dict)  # Modify gui components
    options_dialog = PlotOptionsDialog(options)  # Build dialog w/ options
    if options_dialog.exec() == PlotOptionsDialog.Accepted:
        response_dict = options.values_todict()

    return file_list, data_labels, response_dict


def main(file_list, data_labels, figsize=(6.5, 4.5), basecorrect=True):
    """
    Plot multiple .asc files on one plot with data_labels as legend.

    Parameters
    ----------
    file_list : list of str
        list of full filepaths to ".asc" files for GC data.
    data_labels : list of str
        List of data labels used for generating plot legends.
    figsize : tuple, optional
        Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
        figsize suggestions:
        1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
        1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);
    basecorrect : bool, optional
        Indicates whether or not to baseline correct individual gc data.
        The default is 'True'.

    Returns
    -------
    fig : matplotlib.pyplot.figure
        Figure handle for plot.
    ax : matplotlib.pyplot.axis
        Axis handle for plot.

    """
    fig, ax = plt.subplots()
    plot_tools.set_plot_style(figsize)
    for filename, data_label in zip(file_list, data_labels):
        data = GCData(filename, basecorrect)
        ax.plot(data.time, data.signal, label=data_label)
    plt.legend()
    plt.show()
    return fig, ax


if __name__ == "__main__":
    plt.close('all')
    file_list, data_labels, response_dict = get_user_inputs()
    fig, ax = main(file_list, data_labels, **response_dict)



"""
Prompt user for inputs and run calibration analysis on data set.

Created on Thu Feb  3 09:25:10 2022
@author: Briley Bourgeois
"""
import os
import sys

import pandas as pd
import catalight.analysis.tools as analysis_tools
from catalight.analysis.user_inputs import PlotOptionsDialog, PlotOptionList
from catalight.equipment.experiment_control import Experiment
from PyQt5.QtWidgets import QFileDialog, QApplication


def get_user_inputs(starting_dir=None, cal_folder=None):
    """
    Get inputs from user before running calibration.

    Parameters
    ----------
    starting_dir : str, optional
        Directory to start filedialog in. The default is None.
    cal_folder : str, optional
        Directory containing calibrations. The default is None.

    Returns
    -------
    Experiment
        Experiment object for already run calibration experiment
    pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end, ppm]
    dict :
        figsize, (tuple) Desired size of output figure in inches (x,y).
        forcezero, (bool) Add point (x=0, y=0) to data set.

    """
    app = QApplication(sys.argv)  # noqa
    # Prompt user to select calibration file
    prompt = "Please select the desired calibration file"
    calibration_path = QFileDialog.getOpenFileName(None, prompt, cal_folder,
                                                   "csv file (*.csv)")[0]
    calDF = pd.read_csv(calibration_path, delimiter=',', index_col='Chem ID')

    # Build Expt object from user selected expt directory
    prompt = 'Select Calibration Experiment Folder (containing expt log)'
    expt_dir = QFileDialog.getExistingDirectory(None, prompt)
    log_path = os.path.join(expt_dir, 'expt_log.txt')
    expt = Experiment()  # Initialize experiment obj
    expt.read_expt_log(log_path)  # Read expt parameters from log
    expt.update_save_paths(os.path.dirname(log_path))  # update file paths

    # Edit Options specifically for calibration dialog
    include_dict = {'forcezero': True, 'figsize': True}
    options = PlotOptionList()  # Create default gui options list
    options.change_includes(include_dict)  # Modify gui components
    options_dialog = PlotOptionsDialog(options)  # Build dialog w/ options
    if options_dialog.exec_() == PlotOptionsDialog.Accepted:
        response_dict = options.values_todict()

    return expt, calDF, response_dict


def main(expt, calDF, figsize=(6.5, 4.5), forcezero=True):
    """
    Run analysis.tools.analyze_cal_data. Wrapped w/ Main for consistency.

    Parameters
    ----------
    expt : Experiment
        Experiment object for already run calibration experiment
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end, ppm]
    figsize : tuple, optional
        Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
        figsize suggestions:
        1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
        1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);
    force_zero : bool, optional
        Add point (x=0, y=0) to data set. The default is True.

    Returns
    -------
    matplotlib.axes._axes.Axes
        Axis handle for "run_num" plot
    matplotlib.axes._axes.Axes
        Axis handle for plots showing expected ppm vs measured counts

    """
    run_num_plots, cal_plots = analysis_tools.analyze_cal_data(expt, calDF,
                                                               figsize,
                                                               forcezero)
    return run_num_plots, cal_plots


if __name__ == "__main__":

    expt, calDF, response_dict = get_user_inputs()
    run_num_plots, cal_plots = main(expt, calDF, **response_dict)



"""
Module is used to test the response time of the gas chromatograph.

The GC and gas system are used to check the response time of the system.
The user must make edits to this module in the current iteration. Update the
delay times, flow rates, and gas details to test for. The system will flow gas,
then wait the input delay time until starting the GC. This allows the user to
check how long the it takes for the gas supply to reach steady state to the GC.
For data analysis, see the "run_plot_chromatograms_stacked.py" module.

Created on Tue Dec 13 14:32:19 2022

@author: Briley Bourgeois
"""
import os
import time
from datetime import date

from catalight.equipment.alicat_MFC.gas_control import Gas_System
from catalight.equipment.sri_gc.gc_control import GC_Connector


def main(main_dir, delay_times, flows, gas_list, comp_list_on, comp_list_off,
         ctrl_file=None, flush_flowrate=50):
    """
    Test the response time of the GC by collecting samples in controlled way.

    Loops through flows entered to test the GC response delay time after. The
    composition determined by comp_list_on is sent from the MFCs at the
    supplied flow rate. GC collection is started delay_times after in a loop.
    Data gets saved as in folder based on the flow rate and as files named with
    the delay time.

    Parameters
    ----------
    main_dir : str
        Directory in which experiment directories will be created/saved
    delay_times : list of float or int
        Time between setting comp_list_on and starting GC collection
    flows : list of float or int
        Total gas flow rate in sccm.
    gas_list : list of str
        Input gasses into MFC.
        Example: ['C2H4', 'C2H2', 'H2', 'Ar']
    comp_list_on : list of float or int
        Composition list for test mix.
        Example: [0, 0.1, 0, 0.9] sends 10% Gas B diluted by Gas D
    comp_list_off : list of float or int
        Composition list for flush mix.
        Example: [0, 0, 0, 1] sends 100 gas D, which should be inert gas
    ctrl_file : str
        Full path to the control file for the GC to use. Get entered on
        GC init. The default value is None, which uses the GC class default.
    flush_flowrate : float or int
        [sccm] Total flow rate to use when flushing between gc samples.
        The default is 50 sccm

    Returns
    -------
    None.

    """
    expt_dir = os.path.join(main_dir,
                            (date.today().strftime('%Y%m%d')
                             + '_gc_delay_test'))
    os.makedirs(expt_dir, exist_ok=True)

    gc = GC_Connector(ctrl_file)
    gc.sample_set_size = 1  # Subsequent runs will be wrong delay!!
    gas_control = Gas_System()
    gas_control.set_gasses(gas_list)

    for tot_flow in flows:
        flow_dir = os.path.join(expt_dir, (str(tot_flow) + '_sccm'))
        for delay in delay_times:
            filename = (str(delay) + '_min_delay')
            filepath = os.path.join(flow_dir, filename)
            os.makedirs(filepath, exist_ok=True)
            gc.update_ctrl_file(filepath)  # Changes save path
            gas_control.set_flows(comp_list_on, tot_flow)  # Flow Target Gas

            for second in range(int(delay * 60)):
                time.sleep(1)  # 1 second at a time so script can be stopped

            # Turn on gc and pause a minute to be safe
            gc.peaksimple.SetRunning(1, True)
            time.sleep(60)
            # Clear out gas line
            gas_control.set_flows(comp_list_off, flush_flowrate)

            # Wait until gc collection is done
            for second in range(int((gc.sample_rate - 1) * 60)):
                time.sleep(1)  # 1 second at a time so script can be stopped
    gas_control.disconnect()


if __name__ == "__main__":
    # Example usage:
    # User Inputs:
    # ------------
    main_dir = r'C:\Peak489Win10\GCDATA'  # Save location
    delay_times = [5, 15, 25, 35, 45, 55]  # In minutes
    flows = [5, 25, 50]  # Total gas flow rate in sccm
    gas_list = ['C2H4', 'C2H2', 'H2', 'Ar']  # Input Gasses
    comp_list_on = ([0, 1, 0, 0])  # Composition to use for testing
    comp_list_off = ([0, 0, 0, 1])  # Composition to flush system
    # Function call
    main(main_dir, delay_times, flows,
         gas_list, comp_list_on, comp_list_off)



"""
Graphical user interface classes for helping user input parameters.

This module contains a number of UI classes that prompt the user to select
calibration files, data paths, and enter function parameters.
Created on Thu Jun 23 22:42:55 2022
@author: Briley Bourgeois

References
----------
# noqa
(1) https://github.com/napari/magicgui/issues/9
(2) https://stackoverflow.com/questions/28544425/pyqt-qfiledialog-multiple-directory-selection
(3) https://stackoverflow.com/questions/38746002/pyqt-qfiledialog-directly-browse-to-a-folder
(4) https://stackoverflow.com/questions/4286036/how-to-have-a-directory-dialog
(5) https://stackoverflow.com/questions/38609516/hide-empty-parent-folders-qtreeview-qfilesystemmodel
(6) https://www.youtube.com/watch?v=dqg0L7Qw3ko
(7) https://stackoverflow.com/questions/52592977/how-to-return-variables-from-pyqt5-ui-to-main-function-python
"""
from __future__ import annotations
import os
import sys
from ast import literal_eval
from dataclasses import dataclass, fields, is_dataclass


import catalight.analysis.tools as analysis_tools
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (QDialogButtonBox, QItemDelegate, QDialog,
                             QApplication, QTreeWidget, QAbstractItemView,
                             QGridLayout, QListView, QTreeView, QVBoxLayout,
                             QFileDialog, QFileSystemModel, QTreeWidgetItem,
                             QComboBox, QRadioButton, QLineEdit, QLabel,
                             QDoubleSpinBox)


class DirectorySelector(QFileDialog):
    """
    Open a file dialog that allows selection of multiple directories.

    We use this function instead of regular QFileDialog because
    this lets us select multiple folders. Updates selection mode.

    Parameters
    ----------
    starting_dir : str, optional
        Path in which to start search. The default is None.

    """

    def __init__(self, starting_dir=None):
        super().__init__()
        self.setWindowTitle('Choose Directories')
        self.setGeometry(50, 100, 1800, 800)
        self.setViewMode(1)
        self.setOption(QFileDialog.DontUseNativeDialog, True)
        self.setFileMode(QFileDialog.DirectoryOnly)
        if starting_dir is not None:
            self.setDirectory(starting_dir)

        # This loop cycles through dialog and sets multi-item selection
        for view in self.findChildren((QListView, QTreeView)):
            if isinstance(view.model(), QFileSystemModel):
                view.setSelectionMode(QAbstractItemView.ExtendedSelection)

    def get_output(self):
        """Return list of selected directories."""
        return self.selectedFiles()


@dataclass
class Option():
    """
    Generic option to use when building plotting dialogs.

    See :class:`PlotOptionList` for main usage. Multiple Option items are
    created within :meth:`PlotOptionList.__init__` to take in common user
    inputs. New Option items can be created to easily update GUI components
    for basic UI.
    """
    value: tuple | bool | float | str  #: User supplied value
    include: bool  #: Indicates whether or not to include GUI element
    label: str  #: String for constructing QLabel(label)
    tooltip: str  #: String for widget tooltip
    widget: QComboBox | QDoubleSpinBox | QLineEdit | QRadioButton
    """Widget used for entering option values"""


@dataclass
class PlotOptionList():
    """
    Starting list of options to use when building plotting dialogs.

    General purpose label, tooltips, and default values are already included.
    When utilizing this class, you need to "turn on" GUI elements by setting
    individual Option instances to Option.include = True. This is made easier
    with the change_includes() method.
    """

    _reactant_tip = ('String identity of reactant molecule to track.\nMust '
                     'match what exists in the calibration file exactly.')
    _target_tip = ('String identity of the target to use when calculating '
                   'selectivity.\nMust match what exists in the calibration '
                   'file exacly.')
    _mole_bal_tip = ('Code will perform a mole balance for the element prov'
                     'ided.\nThe default is \'c\'. (i.e. carbon balance)')
    _figsize_tip = ('Desired size of output figure in inches (x,y),\n'
                    'figsize suggestions:\n'
                    '1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);\n'
                    '1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);')

    reactant: Option = Option('', False, 'Reactant', _reactant_tip, QLineEdit)
    """Option: Option instance w/ QLineEdit widget for entering reactant"""

    target_molecule: Option = Option('', False, 'Target Molecule',
                                     _target_tip, QLineEdit)
    """
    Option: Option instance w/ QLineEdit widget for entering target molecule.
    """

    mole_bal: Option = Option('c', False, 'Mole Balance Element',
                              _mole_bal_tip, QLineEdit)
    """Option: Option instance w/ QLineEdit widget for mole balance element."""

    figsize: Option = Option((6.5, 4.5), False, 'Figure Size (x, y)',
                             _figsize_tip, QComboBox)
    """
    Option: Option instance w/ QComboBox widget for entering figure size tuple.
    """

    savedata: Option = Option(True, False, 'Save Data? (T/F)',
                              'Save Data? T/F', QComboBox)
    """Option: Option instance w/ QComboBox widget for entering T/F."""

    switch_to_hours: Option = Option(2, False,
                                     'Time to switch units to hours (hr)',
                                     'Time in hour to switch x axis unit',
                                     QDoubleSpinBox)
    """
    Option: Option instance w/ QDoubleSpinBox widget for entering time
    in which the user would like to switch units from minute to hours.
    """

    overwrite: Option = Option(False, False,
                               'Overwrite previous calculations?',
                               None, QComboBox)
    """Option: Option instance w/ QComboBox widget for entering T/F."""

    basecorrect: Option = Option(True, False, 'Add baseline correction?',
                                 None, QComboBox)
    """Option: Option instance w/ QComboBox widget for entering T/F."""

    xdata: Option = Option('[x1, x2, x3, ...]', False,
                           'Enter array for new X data', 'Alice', QLineEdit)
    """
    Option: Option instance w/ QLineEdit widget for entering new xdata array.
    """

    units: Option = Option('H2/C2H2', False, 'Enter New X Units',
                           None, QLineEdit)
    """Option: Option instance w/ QLineEdit widget for entering X data unit."""

    plot_XandS: Option = Option(False, False,
                                'Plot Conversion and Selectivity (2 plots)',
                                'True to plot conversion and selectivity plots.',  # noqa
                                QRadioButton)
    """Option: Option instance w/ QRadioButton widget for checking T/F."""

    plot_XvsS: Option = Option(False, False,
                               'Plot Selectivity vs Conversion (1 plot)',
                               'True plots selectivity as function of conversion.',  # noqa
                               QRadioButton)
    """Option: Option instance w/ QRadioButton widget for checking T/F."""

    forcezero: Option = Option(True, False, 'Include (0, 0) in fit?',
                               'Add point (x=0, y=0) to data set.',
                               QComboBox)
    """Option: Option instance w/ QLineEdit widget for entering T/F."""

    def __iter__(self):
        """
        If the user hasn't modified the attributes, unpacks individual Options.

        Raises
        ------
        TypeError
            Raises type error if user somehow changes data type of class.

        Returns
        -------
        iterator?
            Returns iterator of instance attributes like unpacking with *tuple.
            If the user hasn't modified the attributes, this will unpack
            individual Option instances.
        """
        if not is_dataclass(self):
            raise TypeError(f"This mixin is dataclass-only,"
                            f"which {type(self)} isn't.")
        return (getattr(self, field.name) for field in fields(self))

    def change_includes(self, option_toggle_dict):
        """
        Set individual options on/off in one command using a dict.

        Parameters
        ----------
        option_toggle_dict : dict
            {field: bool} Use this to set individual
            Options on/off in one command.

        Returns
        -------
        None.

        """
        for key, value in option_toggle_dict.items():
            getattr(self, key).include = value

    def values_todict(self, get_all=False):
        """
        Return dict of values for active Options.

        Parameters
        ----------
        get_all : bool, optional
            Setting to True returns values for hidden GUI elements.
            The default is False.

        Returns
        -------
        values_dict : dict
            {field: value} User supplied values for fields based on GUI entry.

        """

        values_dict = {}
        for field in fields(self):
            key = field.name
            option = getattr(self, key)
            value = option.value

            if option.include or get_all:
                values_dict[key] = value

        return values_dict


class PlotOptionsDialog(QDialog):
    """
    Dynamic dialog window displaying options based on input parameters.

    Supply an instance of PlotOptionList with .include attr turned on for the
    desired GUI elements. When the user presses ok, the PlotOptionList instance
    will have its values updated based on the GUI values provided.

    Parameters
    ----------
    plot_options: PlotOptionList
        Instance of PlotOptionList for which the Option.include attr has been
        updated to True for all desired GUI elements.
    """

    def __init__(self, plot_options):
        super().__init__()
        self.setWindowTitle('Enter Plotting Instructions')

        # Create widgets for window
        self.layout = QGridLayout(self)
        self.options = plot_options
        row_num = 0
        for option in self.options:
            # Skip when Option.include=False
            if not option.include:
                continue

            # Get widgets from Option instance and update properties
            widget = option.widget()  # Instantiate widget
            widget.setToolTip(option.tooltip)
            if isinstance(widget, QComboBox):
                if isinstance(option.value, bool):
                    widget.addItems(['True', 'False'])
                if isinstance(option.value, tuple):
                    widget.addItems(['(4.35, 3.25)', '(5, 3.65)',
                                     '(6.5, 4.5)', '(9, 6.65)'])
                    widget.setEditable(True)
            # Add widgets to layout
            self.layout.addWidget(QLabel(option.label), row_num, 0)
            self.layout.addWidget(widget, row_num, 1)
            option.widget = widget  # Assign instantiated widget back to option
            row_num += 1

        # Create Dialog buttons, connect, and add to window
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok
                                          | QDialogButtonBox.Cancel)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.layout.addWidget(self.buttonBox, row_num, 0, 1, 2, Qt.AlignCenter)

    def accept(self):
        """
        Update Option.field.value for all GUI components included in instance.

        Returns
        -------
        None.

        """
        for option in self.options:
            if not option.include:
                continue
            widget = option.widget
            if isinstance(widget, QComboBox):
                option.value = literal_eval(widget.currentText())
            elif isinstance(widget, QLineEdit):
                option.value = widget.text()
            elif isinstance(widget, QDoubleSpinBox):
                option.value = widget.value()
            elif isinstance(widget, QRadioButton):
                option.value = widget.isChecked()
            else:
                print('widget changed to something incompatible')
        super(PlotOptionsDialog, self).accept()


class MyDelegate(QItemDelegate):
    """Use for sorting items."""

    def createEditor(self, parent, option, index):
        """Create editor."""
        if index.column() == 1:
            return super(MyDelegate, self).createEditor(parent, option, index)
        return None


class DataExtractor(QDialog):
    """
    User interface for selecting data folders.

    Sweeps the starting directory for matching datafiles, then opens a UI
    showing paths matching data_depth. The user can select which paths they'd
    like, and provide a label for that data set using a table and tree widget.

    Parameters
    ----------
    starting_dir : str, optional
        Main directory to initialize gui in. The default is None.
    target : str, optional
        String to identify as "has data". The default is 'avg_conc'.
    suffix : str, optional
        File type to search for. The default is '.csv'.
    data_depth : int, optional
        Depth between data and path to return. The default is 2.
        For example,

        ==========  ==============================================
        data_depth  path returned
        ==========  ==============================================
        0           user/reactions/experiment/Results/avg_conc.csv
        1           user/reactions/experiment/Results
        2           user/reactions/experiment
        ==========  ==============================================

    Returns
    -------
    None.
    """

    def __init__(self, starting_dir=None, target='avg_conc',
                 suffix='.csv', data_depth=2):
        """Initialize extraction gui."""
        # Can be used if this dialog needs to interact with parent
        # super(DataExtractor, self).__init__(parent)
        super().__init__()

        # Define input parameters as attributes
        self.starting_dir = starting_dir
        # String to find partial match with
        self.target = target
        self.suffix = suffix
        # Separation between data and folder to display
        self.data_depth = data_depth

        self.initLayout()
        self.get_expt_dirs()

    def initLayout(self):
        """Initialize layout of GUI."""
        self.setWindowTitle('Select data to be plotted')
        self.setGeometry(50, 50, 1200, 600)
        # Create the treeWidget, populated when get_expt_dirs is run
        delegate = MyDelegate()
        self.treeWidget = QTreeWidget(self)
        self.treeWidget.setItemDelegate(delegate)
        self.treeWidget.setColumnCount(2)
        self.treeWidget.setHeaderLabels(['Name', 'Data Label'])
        self.treeWidget.setAlternatingRowColors(True)

        # create other buttons
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok
                                          | QDialogButtonBox.Cancel)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.cancel)

        # arrange the GUI
        self.gridLayout = QGridLayout()
        self.gridLayout.addWidget(self.buttonBox, 0, 0)

        self.layout = QVBoxLayout(self)
        self.layout.addWidget(self.treeWidget)
        self.layout.addLayout(self.gridLayout)

    def populateTree(self, dataset):
        """
        Display files in a selection gui allowing user to pick relevant files.

        The path shown in the selection tree will be altered by self.data_depth

        Parameters
        ----------
        dataset : list
            list of strings containing paths to data files
        """
        tree_root = (None, {})
        for f in dataset:
            # Append (../*n) to move folder depth desired number of levels
            # Named folder is the folder you want displayed when choosing data
            file_depth_modifier = '/'.join(['..'] * self.data_depth)
            named_folder = os.path.abspath(os.path.join(f,
                                                        file_depth_modifier))
            sample_str = os.path.relpath(named_folder,
                                         os.path.dirname(self.pathRoot))
            parts = sample_str.split('\\')
            node = tree_root

            for i, p in enumerate(parts):
                if p != '':
                    if p not in node[1]:
                        node[1][p] = (QTreeWidgetItem(node[0]), {})
                        # Changing setText to zero keeps name in 'one' column
                        node[1][p][0].setText(0, p)

                        if i == len(parts) - 1:  # If bottom item
                            node[1][p][0].setCheckState(0, Qt.Unchecked)
                            node[1][p][0].setFlags(node[1][p][0].flags()
                                                   | Qt.ItemIsEditable)
                    node = node[1][p]

        # Add top level items
        for node in tree_root[1].values():
            self.treeWidget.addTopLevelItem(node[0])
        self.treeWidget.expandAll()
        for n in range(2):
            self.treeWidget.resizeColumnToContents(n)

    def getParentPath(self, item):
        """Get parent path of item."""
        def getParent(item, outstring):
            if item.parent() is None:
                return outstring
            outstring = item.parent().text(0) + '/' + outstring
            return getParent(item.parent(), outstring)
        output = getParent(item, item.text(0))
        return output

    def get_expt_dirs(self):
        """
        Open file dialog and let user select directories to analyze.

        Then finds files matching target and suffix provided and calls
        populateTree() using those filepaths. Updates pathRoot.

        Returns
        -------
        None.

        """
        selector = DirectorySelector(self.starting_dir)
        if selector.exec_() == QDialog.Accepted:
            expt_dirs = selector.get_output()
        print(expt_dirs)
        filepaths = analysis_tools.list_matching_files(expt_dirs, self.target,
                                                       self.suffix)
        print(filepaths)
        self.pathRoot = os.path.dirname(expt_dirs[0])
        # function populates tree items based on matching criteria specified
        self.populateTree(filepaths)

    def accept(self):
        """Redefine accept button."""
        file_list = []
        data_labels = []
        for item in self.treeWidget.findItems("", Qt.MatchContains
                                              | Qt.MatchRecursive):
            if item.checkState(0) == Qt.Checked:
                # drop end of pathRoot, to not have it twice
                file_list.append(os.path.join(os.path.dirname(self.pathRoot),
                                              self.getParentPath(item)))
                data_labels.append((item.text(1)))
        self._output = (file_list, data_labels)
        print(self._output)
        super(DataExtractor, self).accept()

    def cancel(self):
        """Restart init at picking starting directory."""
        self.get_expt_dirs()

    def get_output(self):
        """
        Get GUI input values.

        Returns
        -------
        (file_list, data_labels) : tuple
            Where file_list is the full path to the files specified by input
            parameters and data_labels is the matching strings provided in the
            GUI.
        """
        return self._output


if __name__ == "__main__":

    app = QApplication(sys.argv)
    main = DataExtractor()
    # main.move(app.desktop().screen().rect().center() - main.rect().center())
    main.show()
    sys.exit(app.exec())
